/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id: DistributedJMXServer.java 16:40:02 ddesjardins $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.communication.jmx;

import java.io.IOException;
import java.util.Hashtable;
import java.util.Set;

import javax.management.InstanceNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanServer;
import javax.management.MBeanServerConnection;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import javax.management.QueryExp;
import javax.management.ReflectionException;

import org.ow2.petals.kernel.server.MBeanHelper;

/**
 * This object is a connection to a JMX server
 * 
 * @author alouis,ddesjardins - eBMWebsourcing
 */
public class RemoteJMXServer {

    /**
     * The Petals JMX Domain
     */
    public static final String PETALS_DOMAIN = "Petals";

    /**
     * Name of the method getServiceDescription
     */
    public static final String GET_SERVICE_DESCRITION = "getServiceDescription";

    /**
     * Name of the method isExchangeWithConsumerOkayForComponent
     */
    public static final String IS_OK_WITH_CONS = "isExchangeWithConsumerOkayForComponent";

    /**
     * Name of the method isExchangeWithProviderOkayForComponent
     */
    public static final String IS_OK_WITH_PROV = "isExchangeWithProviderOkayForComponent";

    /**
     * Name of the method updateNewEndpoints
     */
    public static final String UPDATE_NEW_ENDPOINTS = "updateNewEndpoints";

    /**
     * Connection to the JMX server
     */
    private MBeanServerConnection connection;

    /**
     * JMX local server; Used to simulate a remote connection, for standalone
     * mode
     */
    private MBeanServer localServer;

    /**
     * Creates a new instance of {@link RemoteJMXServer}
     * 
     * @param connector
     */
    public RemoteJMXServer(MBeanServerConnection connection) {
        super();
        this.connection = connection;
    }

    /**
     * Creates a new instance of {@link RemoteJMXServer}
     * 
     * @param localServer
     */
    public RemoteJMXServer(MBeanServer localServer) {
        super();
        this.localServer = localServer;
    }

    /**
     * Retrieve the object name of the admin service
     * 
     * @return object name
     * @throws IOException
     * @throws NullPointerException
     * @throws MalformedObjectNameException
     */
    public ObjectName getAdminServiceMBeanName() throws IOException, MalformedObjectNameException,
            NullPointerException {
        ObjectName result = null;
        Set<?> objNames;
        Hashtable<String, String> attrs = new Hashtable<String, String>();
        attrs.put("name", MBeanHelper.ADMIN_MBEAN);
        attrs.put("type", "service");
        ObjectName objName = new ObjectName(PETALS_DOMAIN, attrs);

        if (connection == null) {
            objNames = localServer.queryNames(objName, null);
        } else {
            objNames = connection.queryNames(objName, null);
        }

        if (objNames != null && objNames.size() == 1) {
            result = (ObjectName) objNames.iterator().next();
        }
        return result;
    }

    /**
     * Retrieve the object name of the admin service
     * 
     * @return object name
     * @throws IOException
     * @throws NullPointerException
     * @throws MalformedObjectNameException
     */
    public ObjectName getEndpointServiceMBeanName() throws IOException,
            MalformedObjectNameException, NullPointerException {
        ObjectName result = null;
        Set<?> objNames;
        Hashtable<String, String> attrs = new Hashtable<String, String>();
        attrs.put("name", MBeanHelper.ENDPOINT_MBEAN);
        attrs.put("type", "service");
        ObjectName objName = new ObjectName(PETALS_DOMAIN, attrs);

        if (connection == null) {
            objNames = localServer.queryNames(objName, null);
        } else {
            objNames = connection.queryNames(objName, null);
        }

        if (objNames != null && objNames.size() == 1) {
            result = (ObjectName) objNames.iterator().next();
        }
        return result;
    }

    /**
     * 
     * @param name
     * @param operationName
     * @param params
     * @param signature
     * @return
     * @throws InstanceNotFoundException
     * @throws MBeanException
     * @throws ReflectionException
     * @throws IOException
     */
    public Object invoke(ObjectName name, String operationName, Object[] params, String[] signature)
            throws InstanceNotFoundException, MBeanException, ReflectionException, IOException {
        Object out = null;
        if (connection == null) {
            out = localServer.invoke(name, operationName, params, signature);
        } else {
            out = connection.invoke(name, operationName, params, signature);
        }

        return out;
    }

    /**
     * 
     * @param name
     * @param queryExp
     * @return
     * @throws InstanceNotFoundException
     * @throws MBeanException
     * @throws ReflectionException
     * @throws IOException
     */
    @SuppressWarnings("unchecked")
    public Set<ObjectName> queryNames(ObjectName name, QueryExp queryExp)
            throws InstanceNotFoundException, MBeanException, ReflectionException, IOException {

        Set<ObjectName> out = null;
        if (connection == null) {
            out = localServer.queryNames(name, queryExp);
        } else {
            out = connection.queryNames(name, queryExp);
        }
        return out;
    }
}
