/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.communication.jndi.agent;

import java.io.Serializable;
import java.util.Hashtable;
import java.util.Map;

import javax.naming.ServiceUnavailableException;

import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.communication.jndi.agent.util.RegistryUtil;
import org.ow2.petals.communication.network.NetworkException;
import org.ow2.petals.communication.network.NetworkService;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.util.LoggingUtil;

/**
 * The distributed implementation of the PEtALS JNDI agent {@link
 * JNDIAgentService}.
 * 
 * @author rnaudin - EbmWebsourcing
 * @since Petals 2.3
 * 
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.communication.jndi.agent.JNDIAgentService.class))
public class PlatformJNDIAgentServiceImpl extends AbstractJNDIAgentServiceImpl {

    /**
     * Logging utilities
     */
    @Monolog(name = "logger")
    private Logger logger;

    /**
     * The Network service fractal component
     */
    @Requires(name = "network", signature = org.ow2.petals.communication.network.NetworkService.class)
    private NetworkService networkService;

    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
    private ConfigurationService configurationService;

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.communication.jndi.agent.AbstractJNDIAgentServiceImpl#
     * containsContext(java.lang.String)
     */
    @Override
    public boolean containsContext(String key) throws ServiceUnavailableException {
        this.log.start(key);

        boolean result = true;
        if (!this.data.containsKey(key)) {
            // try to retrieve the context
            String[] parentContextNames = key.split("/");
            String parentContextName = "/";
            for (int i = 1; i < parentContextNames.length; i++) {
                Map<String, Serializable> parentContext = this.data.get(parentContextName);
                parentContextName = parentContextName + parentContextNames[i] + "/";
                if (parentContext.containsKey(parentContextName)) {
                    this.getContext(parentContextName);
                } else {
                    result = false;
                    break;
                }
            }
        }

        this.log.end();
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.communication.jndi.agent.AbstractJNDIAgentServiceImpl#
     * getContextData(java.lang.String)
     */
    @Override
    public Map<String, Serializable> getContext(String key) throws ServiceUnavailableException {
        this.log.start(key);

        if (!this.data.containsKey(key)) {
            try {
                Map<String, Serializable> context = this.networkService.getReplicatedHashMap(key);
                this.data.put(key, context);
            } catch (NetworkException e) {
                throw new ServiceUnavailableException(
                        "Failed to create replicated Naming Context : " + e.getMessage() + " : "
                                + e.getCause().getMessage());
            }
        }

        this.log.end();

        return this.data.get(key);
    }

    @Override
    public void removeContext(String key) throws ServiceUnavailableException {
        this.log.start(key);

        this.data.remove(key);

        try {
            this.networkService.deleteReplicatedHashMap(key);
        } catch (NetworkException e) {
            throw new ServiceUnavailableException("Failed to delete replicated Naming Context : "
                    + e.getMessage());
        }

        this.log.end();
    }

    /**
     * Initialize the JNDI data
     * 
     * @throws NetworkException
     */
    private void initData() throws NetworkException {
        this.log.start();

        this.data = new Hashtable<String, Map<String, Serializable>>();
        Map<String, Serializable> replicatedHashMap = this.networkService
                .getReplicatedHashMap(INITIAL_CONTEXT);
        this.data.put("/", replicatedHashMap);

        this.log.end();
    }

    @LifeCycle(on = LifeCycleType.START)
    protected void start() {
        this.log = new LoggingUtil(this.logger);
        this.log.start();

        this.domainConfiguration = this.configurationService.getDomainConfiguration();

        try {
            // start the Petals JNDI agent if necessary
            // (no external JNDI server is referenced)
            if (this.domainConfiguration.getJndiFactory() == null) {
                this.initData();
                this.registryUtil = new RegistryUtil(this);
            }
        } catch (Exception e) {
            this.log.error("Failed to start JNDI agent service", e);
        }

        this.log.end();
    }

    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() throws Exception {
        this.log.call();
    }

}
