/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 Ebm Websourcing
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.management.installation;

import java.io.File;
import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import javax.jbi.JBIException;
import javax.management.ObjectName;
import javax.naming.InitialContext;

import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.container.ContainerService;
import org.ow2.petals.container.lifecycle.ComponentLifeCycle;
import org.ow2.petals.container.lifecycle.Installer;
import org.ow2.petals.jbi.descriptor.original.generated.Component;
import org.ow2.petals.jbi.descriptor.original.generated.Component.SharedLibrary;
import org.ow2.petals.jbi.management.admin.AdminServiceMBean;
import org.ow2.petals.jbi.management.task.TaskContextConstants;
import org.ow2.petals.jbi.management.task.installation.InstallationContextConstants;
import org.ow2.petals.jbi.management.task.installation.InstallationTaskFactory;
import org.ow2.petals.jbi.management.util.PackageHelper;
import org.ow2.petals.jbi.messaging.routing.Router;
import org.ow2.petals.kernel.api.server.PetalsException;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.platform.repository.RepositoryService;
import org.ow2.petals.platform.systemstate.SystemStateService;
import org.ow2.petals.processor.TaskProcessor;
import org.ow2.petals.util.LoggingUtil;
import org.ow2.petals.util.ParameterCheckHelper;

/**
 * JBI installation service implementation.
 * 
 * @version $Rev: 154 $ $Date: 2006-03-27 17:30:10 +0200 (lun., 27 mars 2006) $
 * @since Petals 1.0
 * @author <a href="mailto:rmarins@fossilec.com">Rafael Marins</a>
 * @author ofabre - EBM Websourcing
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.jbi.management.installation.InstallationServiceMBean.class))
public class InstallationServiceImpl implements InstallationServiceMBean {

    /**
     * logger wrapper
     */
    protected LoggingUtil log;

    /**
     * Monolog Logger instance.
     */
    @Monolog(name = "logger")
    protected Logger logger;

    /**
     * Package Handler.
     */
    protected PackageHelper packageHandler;

    /**
     * Configuration Service
     */
    @Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
    private ConfigurationService configurationService;

    /**
     * Communication :: JMX Agent Admin.
     */
    @Requires(name = "admin", signature = org.ow2.petals.jbi.management.admin.AdminServiceMBean.class)
    protected AdminServiceMBean adminService;

    /**
     * JBI messaging :: Router Service.
     */
    @Requires(name = "router", signature = org.ow2.petals.jbi.messaging.routing.Router.class)
    protected Router router;

    /**
     * Platform :: Repository Service.
     */
    @Requires(name = "repository", signature = org.ow2.petals.platform.repository.RepositoryService.class)
    protected RepositoryService repositoryService;

    /**
     * Platform :: SystemState Service.
     */
    @Requires(name = "systemstate", signature = org.ow2.petals.platform.systemstate.SystemStateService.class)
    protected SystemStateService systemStateService;

    /**
     * Container :: Container Service.
     */
    @Requires(name = "container", signature = org.ow2.petals.container.ContainerService.class)
    protected ContainerService container;

    protected InitialContext componentInitialContext;

    protected InstallationTaskFactory installationTaskFactory;

    /**
     * Default constructor.
     */
    public InstallationServiceImpl() {

    }

    public Logger getLogger() {
        return this.logger;
    }

    /**
     * @see javax.jbi.management.InstallationServiceMBean#installSharedLibrary(java
     *      .lang.String) Return ERROR_INSTALL_SHARED_LIBRARY if shared library
     *      can't be loaded correctly. It's an implementation choice !
     */
    public synchronized String installSharedLibrary(String slZipURL) {
        this.log.start();

        ParameterCheckHelper.isNullOrEmptyParameterWithLog(slZipURL, "slZipURL", this.log);

        HashMap<String, Object> installationContext = null;
        try {
            // Transform URL to URI
            URI archiveURI = this.packageHandler.getPackageURI(slZipURL, false);

            // Create installation execution context
            installationContext = this.createInstallationContext(archiveURI);

            // Create installation processor and fill it with tasks
            TaskProcessor processor = this.createSLInstallationProcessor(installationContext,
                    archiveURI);

            // Launch the process
            Throwable processResult = processor.process();
            if (processResult != null) {
                throw new Exception("The installation of the Shared Library '" + slZipURL
                        + "' failed.", processResult);
            }
        } catch (Throwable e) {
            this.log.error(e);
            // Catch all throwable to prevent exotic exception type
            // unserializable for external (JMX client) class-loaders
            StringBuilder msg = new StringBuilder(e.getMessage());
            while (e.getCause() != null) {
                e = e.getCause();
                if (e.getMessage() != null) {
                    msg.append("\nCaused by: ").append(e.getMessage());
                }
            }
            throw new RuntimeException(msg.toString());
        }

        this.log.info("Shared Library '"
                + (String) installationContext.get(TaskContextConstants.ENTITY_NAME)
                + "' succesfully installed");

        this.log.end();
        return (String) installationContext.get(TaskContextConstants.ENTITY_NAME);
    }

    /**
     * @see javax.jbi.management.InstallationServiceMBean#loadNewInstaller(java.lang
     *      .String) Return a specific ObjectName if component can't be
     *      installed correctly. It's an implementation choice !
     * @throws RuntimeException
     *             if an error occurs during installation process. We have
     *             chosen this type of exception because, in the specification,
     *             this method doesn't return an exception
     */
    public ObjectName loadNewInstaller(String installZipURL) {
        this.log.start();
        ParameterCheckHelper
                .isNullOrEmptyParameterWithLog(installZipURL, "installZipURL", this.log);
        HashMap<String, Object> installationContext = null;
        try {
            // Transform URL to URI
            URI archiveURI = this.packageHandler.getPackageURI(installZipURL, false);
            // Create installation execution context
            installationContext = this.createInstallationContext(archiveURI);
            // Create installation processor and fill it with tasks
            TaskProcessor processor = this.createComponentInstallationProcessor(
                    installationContext, archiveURI);
            // Launch the installation processor
            Throwable processResult = processor.process();
            if (processResult != null) {
                throw processResult;
            }
        } catch (Throwable e) {
            this.log.error("Unable to install the component " + installZipURL, e);
            // Catch all throwable to prevent exotic exception type
            // unserializable for external (JMX client) class-loaders
            StringBuilder msg = new StringBuilder(e.getMessage());
            while (e.getCause() != null) {
                e = e.getCause();
                if (e.getMessage() != null) {
                    msg.append("\nCaused by: ").append(e.getMessage());
                }
            }
            throw new RuntimeException(msg.toString());
        }
        this.log.info("Component '"
                + (String) installationContext.get(TaskContextConstants.ENTITY_NAME)
                + "' succesfully installed");
        this.log.end();
        return (ObjectName) installationContext
                .get(InstallationContextConstants.INSTALLER_OBJECT_NAME);
    }

    /**
     * @see javax.jbi.management.InstallationServiceMBean#loadInstaller(java.lang
     *      .String)
     */
    public ObjectName loadInstaller(String componentName) {
        this.log.start();

        ParameterCheckHelper
                .isNullOrEmptyParameterWithLog(componentName, "componentName", this.log);

        ObjectName result = this.adminService.getInstallerByName(componentName);

        // Call init on bootstrap
        Installer installer = this.adminService.getInstallerMBeanByName(result);
        if (installer != null) {
            try {
                installer.initBootstrap(false);
            } catch (JBIException e) {
                StringBuilder msg = new StringBuilder("Failed to initialize Component Bootstrap");
                this.log.error(msg.toString(), e);
                msg.append("\nCaused by: ").append(e.getMessage());
                Throwable ex = e.getCause();
                while (ex != null) {
                    if (ex.getMessage() != null) {
                        msg.append("\nCaused by: " + ex.getMessage());
                    }
                    ex = ex.getCause();
                }
                throw new RuntimeException(msg.toString());
            }
        }

        this.log.end();

        return result;
    }

    /**
     * @see javax.jbi.management.InstallationServiceMBean#uninstallSharedLibrary(java.lang.String)
     */
    public boolean uninstallSharedLibrary(String slName) {
        this.log.start();
        boolean result = true;
        HashMap<String, Object> uninstallationContext = null;
        try {
            ParameterCheckHelper.isNullOrEmptyParameterWithLog(slName, "slName", this.log);
            if (hasComponentLinked(slName) == true) {
                result = false;
                log.error("The shared library can't be uninstalled because a component need it");
            } else {
                // Create uninstallation execution context
                uninstallationContext = this.createSLUnInstallationContext(slName);
                // Create uninstallation processor and fill with tasks
                TaskProcessor processor = this
                        .createSLUnInstallationProcessor(uninstallationContext);
                // Launch the process
                Throwable processResult = processor.process();
                if (processResult != null) {
                    result = false;
                    log.error("The uninstallation of the shared library '" + slName + "' failed.",
                            processResult);
                }
            }
        } catch (Throwable e) {
            result = false;
            this.log.error("Unable to uninstall a shared library", e);
        }
        if (result) {
            this.log.info("Shared library '" + slName + "' succesfully uninstalled");
        }
        this.log.end();
        return result;
    }

    /**
     * Check that no component need the given sl
     * 
     * @param slName
     *            the SL identifier
     * @return true if a component at least needs the given SL, false otherwise
     */
    private boolean hasComponentLinked(String slName) {
        Collection<ComponentLifeCycle> componentLifeCycleCollection = null;
        Iterator<ComponentLifeCycle> iterator = null;
        ComponentLifeCycle cycle = null;
        Component description = null;
        List<SharedLibrary> sls = null;
        /*
         * Check for bindings
         */
        componentLifeCycleCollection = adminService.getBindingCompoLifeCycles().values();
        iterator = componentLifeCycleCollection.iterator();
        if (iterator != null) {
            while (iterator.hasNext()) {
                cycle = iterator.next();
                // Retrieve the list of referenced SL
                description = cycle.getComponentDescription();
                sls = description.getSharedLibraryList();
                if (sls != null) {
                    for (SharedLibrary sl : sls) {
                        if (slName.equals(sl.getContent())) {
                            return true;
                        }
                    }
                }
            }
        }
        /*
         * Check for engines
         */
        componentLifeCycleCollection = adminService.getEngineCompoLifeCycles().values();
        iterator = componentLifeCycleCollection.iterator();
        if (iterator != null) {
            while (iterator.hasNext()) {
                cycle = (ComponentLifeCycle) iterator.next();
                // Retrieve the list of referenced SL
                description = cycle.getComponentDescription();
                sls = description.getSharedLibraryList();
                if (sls != null) {
                    for (SharedLibrary sl : sls) {
                        if (slName.equals(sl.getContent())) {
                            return true;
                        }
                    }
                }
            }
        }
        return false;
    }

    /**
     * TODO : isToBeDeleted parameter isn't used for the moment. This method
     * deletes entirely the component from the container
     * 
     * @see javax.jbi.management.InstallationServiceMBean#unloadInstaller(java.lang
     *      .String, boolean)
     */
    public boolean unloadInstaller(String componentName, boolean isToBeDeleted) {
        this.log.start();

        ParameterCheckHelper
                .isNullOrEmptyParameterWithLog(componentName, "componentName", this.log);

        boolean result = true;

        HashMap<String, Object> uninstallationContext = null;
        try {
            // Create uninstallation execution context
            uninstallationContext = this.createComponentUnInstallationContext(componentName);

            // Create uninstallation processor and fill it with tasks
            TaskProcessor processor = this
                    .createComponentUnInstallationProcessor(uninstallationContext);

            // Launch the process
            Throwable processResult = processor.process();
            if (processResult != null) {
                throw new Exception("The uninstallation of the component '" + componentName
                        + "' failed.", processResult);
            }
            this.log.info("Component '" + componentName + "' succesfully uninstalled");
        } catch (Throwable e) {
            this.log.error("Unable to unload the installer: ", e);
            result = false;
        }
        this.log.end();
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.jbi.management.service.InstallationServiceMBean#
     * getComponentsForInstalledSharedlibrary(java.lang.String)
     */
    public String[] retrieveComponentsForInstalledSharedlibrary(String slName)
            throws PetalsException {
        throw new PetalsException("Not implemented");
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.jbi.management.service.InstallationServiceMBean#
     * getinstalledSharedLibrary()
     */
    public String[] retrieveInstalledSharedLibraries() throws PetalsException {

        return this.systemStateService.getAllInstalledSharedLibNames().toArray(new String[0]);
    }

    /**
     * Shutdown the InstallationServiceMBean. All shared libraries and
     * components are uninstalled
     */
    public void shutdown() throws Exception {
        this.unloadAllComponentInstallers(true);

        this.uninstallAllSharedLibraries();
    }

    /**
     * Start all the components
     * 
     * @return the list of components that have been started
     */
    public String[] startAllComponents() {
        List<String> startedComponents = new ArrayList<String>();

        for (String name : this.systemStateService.getAllLoadedComponentNames()) {
            try {
                this.startComponent(name);
                startedComponents.add(name);
            } catch (JBIException e) {
                this.log.warning("The following component can't be started: " + name + ". Cause: "
                        + e.getMessage(), e);
            }
        }

        return startedComponents.toArray(new String[startedComponents.size()]);
    }

    /**
     * Stop all the components
     * 
     * @return the list of components that have been stopped
     */
    public String[] stopAllComponents() {
        List<String> stopedComponents = new ArrayList<String>();

        for (String name : this.systemStateService.getAllLoadedComponentNames()) {
            try {
                this.stopComponent(name);
                stopedComponents.add(name);
            } catch (JBIException e) {
                this.log.warning("The following component can't be stoped: " + name + ". Cause: "
                        + e.getMessage(), e);
            }
        }
        return stopedComponents.toArray(new String[stopedComponents.size()]);
    }

    /**
     * Shutdown all the components
     * 
     * @return the list of the components that have been shutdowned
     */
    public String[] shutdownAllComponents() {
        List<String> shutdownComponents = new ArrayList<String>();

        for (String name : this.systemStateService.getAllLoadedComponentNames()) {
            try {
                this.shutdownComponent(name);
                shutdownComponents.add(name);
            } catch (JBIException e) {
                this.log.warning("The following component can't be shutdown: " + name + ". Cause: "
                        + e.getMessage(), e);
            }
        }
        return shutdownComponents.toArray(new String[shutdownComponents.size()]);
    }

    /**
     * Start the component
     * 
     * @param componentName
     *            the component to start
     * @return true if the component has been started
     * @throws JBIException
     *             if the component can not be found
     */
    public boolean startComponent(String componentName) throws JBIException {
        ParameterCheckHelper
                .isNullOrEmptyParameterWithLog(componentName, "componentName", this.log);

        ComponentLifeCycle componentLifeCycle = this.adminService
                .getComponentLifeCycleByName(componentName);

        boolean result = false;

        if (componentLifeCycle != null) {
            componentLifeCycle.start();
            result = true;
        } else {
            throw new JBIException("The requested component isn't installed: " + componentName);
        }

        return result;
    }

    /**
     * Stop a component
     * 
     * @param componentName
     *            the component to stop
     * 
     * @return true if the component has been stopped
     * @throws JBIException
     *             if the component can not be found
     */
    public boolean stopComponent(String componentName) throws JBIException {
        ParameterCheckHelper
                .isNullOrEmptyParameterWithLog(componentName, "componentName", this.log);
        ComponentLifeCycle componentLifeCycle = this.adminService
                .getComponentLifeCycleByName(componentName);
        boolean result = false;
        if (componentLifeCycle != null) {
            componentLifeCycle.stop();
            result = true;
        } else {
            throw new JBIException("The requested component isn't installed: " + componentName);
        }
        return result;
    }

    /**
     * Shut down a component
     * 
     * @param componentName
     *            the component to shutdown
     * @return true if the component has been shut down
     * @throws JBIException
     *             if the component has not been found
     */
    public boolean shutdownComponent(String componentName) throws JBIException {
        ParameterCheckHelper
                .isNullOrEmptyParameterWithLog(componentName, "componentName", this.log);

        ComponentLifeCycle componentLifeCycle = this.adminService
                .getComponentLifeCycleByName(componentName);

        boolean result = false;

        if (componentLifeCycle != null) {
            componentLifeCycle.shutDown();
            result = true;
        } else {
            throw new JBIException("The requested component isn't installed: " + componentName);
        }

        return result;
    }

    /**
     * Uninstall all the components
     * 
     * @return the list of components that have been stopped
     */
    public String[] uninstallAllComponents() {
        List<String> uninstalledComponents = new ArrayList<String>();

        for (String name : this.systemStateService.getAllLoadedComponentNames()) {
            try {
                this.uninstallComponent(name);
                uninstalledComponents.add(name);
            } catch (JBIException e) {
                this.log.warning("The following component can't be uninstalled: " + name
                        + ". Cause: " + e.getMessage(), e);
            }
        }
        return uninstalledComponents.toArray(new String[uninstalledComponents.size()]);
    }

    /**
     * Uninstall a component
     * 
     * @param componentName
     *            the component to uninstall
     * 
     * @return true if the component has been uninstalled
     * @throws JBIException
     *             if the component can not be found
     */
    public boolean uninstallComponent(String componentName) throws JBIException {
        ParameterCheckHelper
                .isNullOrEmptyParameterWithLog(componentName, "componentName", this.log);

        ObjectName instName = this.loadInstaller(componentName);
        Installer installer = this.adminService.getInstallerMBeanByName(instName);

        boolean result = false;

        if (installer != null) {
            installer.uninstall();
            result = true;
        } else {
            throw new JBIException("The requested component isn't loaded: " + componentName);
        }

        return result;
    }

    /**
     * Uninstall all the components
     * 
     * @return the list of uninstalled components
     */
    public String[] unloadAllComponentInstallers(boolean force) {

        List<String> removedComponents = new ArrayList<String>();
        for (String compName : this.systemStateService.getAllLoadedComponentNames()) {
            if (force) {
                if (this.forceUnloadComponentInstaller(compName)) {
                    removedComponents.add(compName);
                }
            } else {
                if (this.unloadInstaller(compName, true)) {
                    removedComponents.add(compName);
                } else {
                    this.log.warning("A component can't be correctly uninstalled: " + compName);
                }
            }
        }

        return removedComponents.toArray(new String[removedComponents.size()]);
    }

    /**
     * Uninstall all the shared libraries
     * 
     * @return the list of uninstalled shared libraries
     */
    public String[] uninstallAllSharedLibraries() {
        List<String> removedSharedLibraries = new ArrayList<String>();
        for (String slName : this.systemStateService.getAllInstalledSharedLibNames()) {
            if (this.uninstallSharedLibrary(slName)) {
                removedSharedLibraries.add(slName);
            } else {
                this.log.warning("A shared library can't be correctly uninstalled: " + slName);
            }
        }
        return removedSharedLibraries.toArray(new String[removedSharedLibraries.size()]);
    }

    /**
     * Force to unload installer of the given component
     * 
     * @param componentName
     *            the component to uninstall
     */
    public boolean forceUnloadComponentInstaller(String componentName) {
        boolean removed = false;

        try {
            this.tryToShutDownComponent(componentName);
            this.tryToUninstallComponent(componentName);
            removed = this.unloadInstaller(componentName, true);
        } catch (JBIException e) {
            this.log.warning("Force uninstall on the following component failed: " + componentName,
                    e);
        }

        return removed;
    }

    /**
     * Initialize the InstallationServiceImpl
     */
    @LifeCycle(on = LifeCycleType.START)
    protected void start() {
        this.log = new LoggingUtil(this.logger);
        this.log.start();

        // Initialize package handler
        this.packageHandler = new PackageHelper(this.logger, this.repositoryService);

        String installationDirectory = this.configurationService.getContainerConfiguration()
                .getInstallDirectoryPath();

        // Initialize installation task factory
        this.installationTaskFactory = new InstallationTaskFactory(this.packageHandler,
                this.systemStateService, this.repositoryService, this.log, this.logger,
                this.adminService, this.router, this.container, installationDirectory);

        // Clean the success file in the work directory
        File workFile = new File(installationDirectory, "work");
        for (File file : workFile.listFiles()) {
            if (!file.isDirectory() && file.getName().endsWith(".success")) {
                file.delete();
            }
        }

        this.log.end();
    }

    /**
     * Stop the InstallationServiceImpl
     */
    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() {
        this.log.call();
    }

    private void tryToUninstallComponent(String componentName) throws JBIException {
        ObjectName instName = this.loadInstaller(componentName);
        Installer installer = this.adminService.getInstallerMBeanByName(instName);
        if (installer != null) {
            if (installer.isInstalled()) {
                installer.uninstall();
            }
        } else {
            throw new JBIException("The installer of the component '" + componentName
                    + "' is not loaded");
        }
    }

    private void tryToShutDownComponent(String componentName) throws JBIException {
        ComponentLifeCycle componentLifeCycle = this.adminService
                .getComponentLifeCycleByName(componentName);
        if (componentLifeCycle != null) {
            String compState = componentLifeCycle.getCurrentState();
            if (javax.jbi.management.LifeCycleMBean.STARTED.equals(compState)) {
                this.stopComponent(componentName);
                this.shutdownComponent(componentName);
            } else if (javax.jbi.management.LifeCycleMBean.STOPPED.equals(compState)) {
                this.shutdownComponent(componentName);
            }
        }
    }

    /**
     * Create the installation context
     * 
     * @param zipUrl
     *            String containing the location URL of the Component or Shared
     *            lib ZIP file
     * @return the Installation Context HashMap
     */
    private HashMap<String, Object> createInstallationContext(URI archiveURI) {
        HashMap<String, Object> installationContext = new HashMap<String, Object>();
        installationContext.put(TaskContextConstants.ARCHIVE_URI, archiveURI);

        return installationContext;
    }

    /**
     * Create the shared lib uninstallation context
     * 
     * @param name
     *            name of the shared lib to uninstall
     * @return the Uninstallation Context HashMap
     */
    private HashMap<String, Object> createSLUnInstallationContext(String name) {
        HashMap<String, Object> uninstallationContext = new HashMap<String, Object>();
        uninstallationContext.put(TaskContextConstants.ENTITY_NAME, name);
        return uninstallationContext;
    }

    /**
     * Create the installation processor and add all required task for the
     * complete installation of a component
     * 
     * @param installationContext
     *            the installation context
     * @param archiveURI
     *            the component archive URI
     * @return the filled installation processor
     */
    private TaskProcessor createComponentInstallationProcessor(
            HashMap<String, Object> installationContext, URI archiveURI) {
        TaskProcessor processor = new TaskProcessor(installationContext, this.log);
        String scheme = archiveURI.getScheme();
        if (scheme.equalsIgnoreCase("http") || scheme.equalsIgnoreCase("ftp")) {
            processor.addTask(this.installationTaskFactory.getDownloadTask());
        } else if (scheme.equalsIgnoreCase("file")) {
            processor.addTask(this.installationTaskFactory.getCopyArchiveToWorkDirTask());
        } else {
            throw new IllegalArgumentException("URL protocol not supported:" + scheme);
        }
        processor.addTask(this.installationTaskFactory.getExtractTask());
        processor.addTask(this.installationTaskFactory.getComponentCheckPackageTask());
        processor.addTask(this.installationTaskFactory.getCheckInstalledSLForCompTask());
        processor.addTask(this.installationTaskFactory.getCreateComponentFolderTask());
        processor.addTask(this.installationTaskFactory.getCreateComponentStateHolderTask());
        processor.addTask(this.installationTaskFactory.getCreateAndRegisterInstallerMBeanTask());
        processor.addTask(this.installationTaskFactory.getCopyArchiveToInstalledDirTask());
        processor.addTask(this.installationTaskFactory.getSuccessTask());
        return processor;
    }

    /**
     * Create the installation processor and add all required task for the
     * complete installation of a shared lib
     * 
     * @param installationContext
     *            the installation context
     * @param archiveURI
     *            the sl archive URI
     * @return the filled installation processor
     */
    private TaskProcessor createSLInstallationProcessor(
            HashMap<String, Object> installationContext, URI archiveURI) {
        TaskProcessor processor = new TaskProcessor(installationContext, this.log);
        String scheme = archiveURI.getScheme();
        if (scheme.equalsIgnoreCase("http") || scheme.equalsIgnoreCase("ftp")) {
            processor.addTask(this.installationTaskFactory.getDownloadTask());
        } else if (scheme.equalsIgnoreCase("file")) {
            processor.addTask(this.installationTaskFactory.getCopyArchiveToWorkDirTask());
        } else {
            throw new IllegalArgumentException("URL protocol not supported:" + scheme);
        }
        processor.addTask(this.installationTaskFactory.getExtractTask());
        processor.addTask(this.installationTaskFactory.getSLCheckPackageTask());
        processor.addTask(this.installationTaskFactory.getCopySLPackageToRepositoryTask());
        processor.addTask(this.installationTaskFactory.getCreateSLStateHolderTask());
        processor.addTask(this.installationTaskFactory.getCreateSLClassloaderTask());
        processor.addTask(this.installationTaskFactory.getCopyArchiveToInstalledDirTask());
        processor.addTask(this.installationTaskFactory.getSuccessTask());
        return processor;
    }

    /**
     * Create the uninstallation processor and add all required task for the
     * complete uninstallation of a component
     * 
     * @param uninstallationContext
     *            the uninstallation context
     * @return the filled uninstallation processor
     */
    private TaskProcessor createComponentUnInstallationProcessor(
            HashMap<String, Object> uninstallationContext) {
        TaskProcessor processor = new TaskProcessor(uninstallationContext, log);
        processor.addTask(this.installationTaskFactory.getUnregisterInstallerTask());
        processor.addTask(this.installationTaskFactory.getRemoveComponentStateHolderTask());
        processor.addTask(this.installationTaskFactory.getRemoveComponentFolderTask());
        processor.addTask(this.installationTaskFactory.getRemoveComponentArchiveTask());
        return processor;
    }

    /**
     * Create the uninstallation processor and add all required task for the
     * complete uninstallation of a shared lib
     * 
     * @param uninstallationContext
     *            the uninstallation context
     * @return the filled uninstallation processor
     */
    private TaskProcessor createSLUnInstallationProcessor(
            HashMap<String, Object> uninstallationContext) {
        TaskProcessor processor = new TaskProcessor(uninstallationContext, this.log);
        processor.addTask(this.installationTaskFactory.getRemoveSLClassloaderTask());
        processor.addTask(this.installationTaskFactory.getRemoveSLStateHolderTask());
        processor.addTask(this.installationTaskFactory.getRemoveSLFolderTask());
        processor.addTask(this.installationTaskFactory.getRemoveSLArchiveTask());
        return processor;
    }

    /**
     * Create the component uninstallation context
     * 
     * @param name
     *            name of the component to uninstall
     * @return the Uninstallation Context HashMap
     */
    private HashMap<String, Object> createComponentUnInstallationContext(String name) {
        HashMap<String, Object> uninstallationContext = new HashMap<String, Object>();
        uninstallationContext.put(TaskContextConstants.ENTITY_NAME, name);
        return uninstallationContext;
    }

}
