/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.management.task;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.util.HashMap;
import java.util.concurrent.atomic.AtomicLong;

import org.apache.commons.io.FileUtils;
import org.ow2.petals.kernel.api.server.PetalsException;
import org.ow2.petals.util.LoggingUtil;
import org.ow2.petals.util.ZipUtil;

/**
 * This task extracts entity (component, shared lib, service assembly) archive
 * into the work directory <br/>
 * Inputs : archive {@link URI} <br/>
 * Outputs : {@link File} pointing on unzipped archive <br/>
 * The task has got a non-robust approach : it stops at the first subtask failed<br />
 * and have the undo capability.<br />
 * 
 * @author ofabre - EBM Websourcing
 * @author Frederic Gardes - eBM Websourcing
 */
public class ExtractTask extends AbstractFileManipulationTask {

    /**
     * The temporary unique id suffix
     */
    private static AtomicLong uniqueIdSuffix = new AtomicLong();

    public ExtractTask(LoggingUtil log, String petalsInstallDirectory) {
        super(log, petalsInstallDirectory);
    }

    @SuppressWarnings("unchecked")
    public void execute(HashMap context) throws Exception {
        URI archiveURI = (URI) context.get(TaskContextConstants.ARCHIVE_URI);
        // Explode Comp
        File unzipRoot = explodeEntityPackage(archiveURI);
        // Fill context
        context.put(TaskContextConstants.UNZIP_ROOT, unzipRoot);
    }

    private File explodeEntityPackage(URI archiveURI) throws PetalsException {
        File zipFile = new File(archiveURI);
        File destFile = new File(getWorkDirectory(), zipFile.getName() + "-"
                + uniqueIdSuffix.incrementAndGet());
        try {
            if (!destFile.mkdirs()) {
                throw new PetalsException("Failed to create temporary work dir: "
                        + destFile.getAbsolutePath());
            }
            ZipUtil.unzipArchive(zipFile, destFile);
        } catch (IOException e) {
            try {
                FileUtils.forceDelete(destFile);
            } catch (IOException e1) {
                // Nothing to do, just log
                log.warning("Residual incomplete component folder");
            }
            throw new PetalsException(e);
        }
        return destFile;
    }

    @SuppressWarnings("unchecked")
    public void undo(HashMap context) throws Exception {
        if (context.containsKey(TaskContextConstants.UNZIP_ROOT)) {
            File tempDir = (File) context.get(TaskContextConstants.UNZIP_ROOT);
            if (tempDir != null) {
                try {
                    if (tempDir.exists()) {
                        FileUtils.forceDelete(tempDir);
                    }
                } catch (IOException e) {
                    log.error("Failed to revert a ExtractTask: " + e.getMessage());
                }
            }
        }
    }

}
