/**
 * PETALS - PETALS Services Platform. Copyright (c) 2006 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id: ExtractSUsTask.java 154 6 oct. 06 ofabre $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.management.task.deployment.deploy;

import java.net.URI;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.jbi.descriptor.original.generated.ServiceUnit;
import org.ow2.petals.jbi.management.ManagementException;
import org.ow2.petals.jbi.management.task.AbstractLoggableTask;
import org.ow2.petals.jbi.management.task.TaskContextConstants;
import org.ow2.petals.jbi.management.task.deployment.DeploymentContextConstants;
import org.ow2.petals.jbi.management.task.deployment.DeploymentUtils;
import org.ow2.petals.jbi.management.util.PackageHelper;
import org.ow2.petals.kernel.api.server.PetalsException;
import org.ow2.petals.platform.repository.RepositoryService;
import org.ow2.petals.util.LoggingUtil;

/**
 * This task extracts service unit archives into the repository <br/>
 * Inputs : {@link Jbi}, parent service assembly install root {@link URI} <br/>
 * Outputs : SU install roots {@link Map} <br/>
 * The task has got a non-robust approach : it stops at the first subtask failed<br />
 * and have the undo capability.<br />
 * 
 * @author ofabre - EBM Websourcing
 * @author Frederic Gardes - eBM Websourcing
 */
public class ExtractSUsTask extends AbstractLoggableTask {

    /**
     * Package Handler
     */
    protected PackageHelper packageHandler;

    /**
     * Platform Component :: Repository Service
     */
    protected RepositoryService repositoryService;

    public ExtractSUsTask(LoggingUtil log, PackageHelper packageHandler,
            RepositoryService repositoryService) {
        super(log);
        this.packageHandler = packageHandler;
        this.repositoryService = repositoryService;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.processor.Task#execute(java.util.HashMap)
     */
    public void execute(HashMap<String, Object> context) throws Exception {
        Jbi descriptor = (Jbi) context.get(TaskContextConstants.DESCRIPTOR);
        URI saInstallRoot = (URI) context.get(TaskContextConstants.INSTALL_ROOT);
        /*
         * Explodes SUs
         */
        Map<String, URI> suInstallRoots = expendSUS(saInstallRoot, descriptor);
        /*
         * Fill context
         */
        context.put(DeploymentContextConstants.SU_INSTALL_ROOTS, suInstallRoots);
    }

    /**
     * Explode service unit packages into the repository
     * 
     * @param saInstallRoot
     *            the root {@link URI} of the service assembly
     * @param descriptor
     *            the service assembly descriptor
     * @return the map containing all su install roots
     * @throws ManagementException
     *             if the repository service doesn't manage to expend SU archive
     * @throws PetalsException
     */
    protected Map<String, URI> expendSUS(URI saInstallRoot, Jbi descriptor)
            throws ManagementException, PetalsException {
        Map<String, URI> suInstallRoots = new HashMap<String, URI>();
        List<ServiceUnit> sus = descriptor.getServiceAssembly().getServiceUnit();
        for (ServiceUnit unit : sus) {
            URI suArchiveURI = getSUArchiveURI(unit, saInstallRoot);
            URI suInstallRoot = expandSUIntoSA(unit.getIdentification().getName(), suArchiveURI,
                    DeploymentUtils.getServiceAssemblyName(descriptor));
            suInstallRoots.put(unit.getIdentification().getName(), suInstallRoot);
        }
        return suInstallRoots;
    }

    /**
     * Retrieve the service unit archive {@link URI}
     * 
     * @param su
     *            the {@link ServiceUnit} production element
     * @param saInstallRoot
     *            the sa installation root {@link URI}
     * @return the su archive {@link URI}
     * @throws PetalsException
     */
    protected URI getSUArchiveURI(ServiceUnit su, URI saInstallRoot) throws PetalsException {
        String suZipNAme = su.getTarget().getArtifactsZip().trim();
        String suZipURL = "file://" + saInstallRoot.getPath() + suZipNAme;
        return packageHandler.getPackageURI(suZipURL.replaceAll(" ", "%20"), true);
    }

    /**
     * Expand a service unit package into its service assembly installRoot.
     * 
     * @param suId
     *            the unique identifier of the su
     * @param suZipLocation
     *            URI representing the su zip location
     * @param saId
     *            the unique identifier of the sa
     * @return SU installationRoot, null if it fails
     * @throws ManagementException
     *             if it fails to expend the su package to the repository
     */
    protected URI expandSUIntoSA(String suId, URI suZipLocation, String saId)
            throws ManagementException {
        URI result = null;
        try {
            result = repositoryService.explodeSUIntoSAInstallFolder(suId, suZipLocation, saId);
        } catch (PetalsException pe) {
            throw new ManagementException("Unexpected error expending service JBI installation "
                    + "package into repository", pe);
        }
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.processor.Task#undo(java.util.HashMap)
     */
    public void undo(HashMap<String, Object> context) throws Exception {
        if (context.containsKey(TaskContextConstants.DESCRIPTOR)
                && context.containsKey(TaskContextConstants.INSTALL_ROOT)) {
            Jbi descriptor = (Jbi) context.get(TaskContextConstants.DESCRIPTOR);
            URI saInstallRoot = (URI) context.get(TaskContextConstants.INSTALL_ROOT);
            if (descriptor != null && saInstallRoot != null) {
                removeSA(DeploymentUtils.getServiceAssemblyName(descriptor));
            }
        }
        /*
         * Fill context
         */
        context.remove(DeploymentContextConstants.SU_INSTALL_ROOTS);
    }

    /**
     * Remove the service assembly into the repository
     * 
     * @param serviceAssemblyName
     *            Identifiant of the service assembly
     */
    private void removeSA(String serviceAssemblyName) {
        try {
            repositoryService.removeServiceAssemblyPackage(serviceAssemblyName);
        } catch (PetalsException pe) {
            log.warning("Unable to remove an su in the repository", pe);
        }
    }

}
