/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id: EndpointRegistry.java 11:27:27 ddesjardins $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.messaging.registry;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.Callable;
import java.util.concurrent.CompletionService;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.naming.Binding;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.LinkRef;
import javax.naming.Name;
import javax.naming.NameAlreadyBoundException;
import javax.naming.NameNotFoundException;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.NotContextException;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.easywsdl.extensions.wsdl4complexwsdl.WSDL4ComplexWsdlFactory;
import org.ow2.easywsdl.extensions.wsdl4complexwsdl.api.WSDL4ComplexWsdlReader;
import org.ow2.easywsdl.wsdl.WSDLFactory;
import org.ow2.easywsdl.wsdl.api.Description;
import org.ow2.easywsdl.wsdl.api.InterfaceType;
import org.ow2.easywsdl.wsdl.api.Service;
import org.ow2.easywsdl.wsdl.api.WSDLException;
import org.ow2.easywsdl.wsdl.api.WSDLReader;
import org.ow2.petals.communication.jmx.JMXService;
import org.ow2.petals.communication.jmx.RemoteJMXServer;
import org.ow2.petals.communication.jndi.client.JNDIService;
import org.ow2.petals.jbi.descriptor.original.generated.LinkType;
import org.ow2.petals.jbi.management.admin.AdminServiceMBean;
import org.ow2.petals.jbi.messaging.endpoint.AbstractEndpoint;
import org.ow2.petals.jbi.messaging.endpoint.ConsumerEndpoint;
import org.ow2.petals.jbi.messaging.endpoint.ExternalEndpoint;
import org.ow2.petals.jbi.messaging.endpoint.InternalEndpoint;
import org.ow2.petals.jbi.messaging.endpoint.LinkedEndpoint;
import org.ow2.petals.kernel.api.server.PetalsException;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.util.LoggingUtil;
import org.ow2.petals.util.NamingHelper;
import org.ow2.petals.util.ParameterCheckHelper;
import org.w3c.dom.Document;

import static org.ow2.petals.kernel.api.Constants.UNRESOLVED_INTERFACE;

/**
 * This registry is the registry used in a distributed environment. It used a
 * JNDI distributed directory to manage the endpoint storage. <br>
 * Only subdomain-endpoints or public endpoints from other subdomains are
 * retrieved by the getXXX() methods
 * 
 * This is the JNDIRegistry service format :
 * 
 * <pre>
 * '/'
 *  '/endpoints/'
 *  '          /e1' (key is serviceName@endpointName)
 *  '          /e2'
 *  '/services/'
 *  '         /s1/ref-e1' (key is serviceName and object is link to /endpoints/e1)
 *  '         /s2/ref-e2'
 *  '/interfaces/'
 *  '           /i1/ref-e1' (key is itfName and object is link to /endpoints/e1)
 *  '           /i2/ref-e2'
 *  '/new-endpoints/'
 *  '              /ref-e1' (link to /endpoints/e1)
 *  '              /ref-e2' (link to /endpoints/e2)
 * </pre>
 * 
 * @author alouis - ddesjardins - chamerling - rnaudin - EBM WebSourcing
 * @author cdeneux - Cap Gemini
 * @since Petals 1.1
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.jbi.messaging.registry.EndpointRegistry.class))
public class EndpointRegistryImpl implements EndpointRegistry {

    /**
     * Logger wrapper
     */
    protected LoggingUtil log;

    /**
     * Monolog Logger instance.
     */
    @Monolog(name = "logger")
    protected Logger logger;

    /**
     * Manager of the distributed JMX server
     */
    @Requires(name = "jmx", signature = org.ow2.petals.communication.jmx.JMXService.class)
    protected JMXService jmxService;

    /**
     * JNDI service fractal component
     */
    @Requires(name = "jndi", signature = org.ow2.petals.communication.jndi.client.JNDIService.class)
    protected JNDIService jndiService;

    /**
     * Admin service fractal component
     */
    @Requires(name = "admin", signature = org.ow2.petals.jbi.management.admin.AdminServiceMBean.class)
    protected AdminServiceMBean adminService;

    /**
     * Configuration service fractal component
     */
    @Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
    protected ConfigurationService configurationService;

    /**
     * The local container name.
     */
    private String containerName;

    /**
     * Subdomain name
     */
    private String subdomainName;

    /**
     * Name of the JNDI endpoints context
     */
    public static final String ENDPOINTS_REF = "endpoints";

    /**
     * Name of the JNDI interfaces context
     */
    public static final String INTERFACES_REF = "interfaces";

    /**
     * Name of the JNDI new endpoints context
     */
    public static final String NEW_ENDPOINTS_REF = "new-endpoints";

    /**
     * Name of the JNDI service context
     */
    public static final String SERVICES_REF = "services";

    /**
     * WSDL reader
     */
    protected WSDL4ComplexWsdlReader wsdlReader;

    /**
     * Endpoint context
     */
    protected Context endpointsContext;

    /**
     * Interfaces context
     */
    protected Context interfacesContext;

    /**
     * New endpoint context
     */
    protected Context newEndpointContext;

    /**
     * Root context
     */
    protected InitialContext rootContext;

    /**
     * Service context
     */
    protected Context servicesContext;

    /**
     * The timer to update the new endpoints in background
     */
    private Timer updateLocalNewEndpointsTimer;

    /**
     * The Flag to update the new endpoints in background
     */
    private AtomicBoolean updateScheduled;

    /**
     * The Completion Service to process the updating of new endpoint
     * asynchronously
     */
    private CompletionService<Object> updateNewEndpointsCS;

    /**
     * The Thread Pool Executor to process the updating of new endpoint
     * asynchronously
     */
    private ThreadPoolExecutor updateNewEndpointsTPE;

    /**
     * Inner task which update the local new endpoints in background.
     * 
     */
    private class UpdateLocalNewEndpointsTask extends TimerTask {

        @Override
        public void run() {
            EndpointRegistryImpl.this.log.debug("Updating local new endpoints in background...");

            try {
                EndpointRegistryImpl.this.updateNewEndpoints(false);
            } catch (final JBIException e) {
                EndpointRegistryImpl.this.log.error(e);
            }

            EndpointRegistryImpl.this.updateScheduled.set(false);
        }

    }

    /**
     * Inner future which update a local new endpoint.
     * 
     */
    private class UpdateLocalNewEndpointFuture implements Callable<Object> {

        private final AbstractEndpoint endpoint;

        public UpdateLocalNewEndpointFuture(AbstractEndpoint endpoint) {
            this.endpoint = endpoint;
        }

        public Object call() throws Exception {
            EndpointRegistryImpl.this.log.debug("Updating a new endpoint : "
                    + this.endpoint.getEndpointName());

            EndpointRegistryImpl.this.updateEndpointDescriptionAndInterfaces(this.endpoint);
            try {
                EndpointRegistryImpl.this.validateEndpoint(this.endpoint);
            } catch (final JBIException e) {
                EndpointRegistryImpl.this.log.error("Problem while updating new endpoint '"
                        + this.endpoint.getEndpointName() + "'", e);
            }

            return null;
        }

    }

    /**
     * Inner future which call for update of remote new endpoints.
     * 
     */
    private class UpdateRemoteNewEndpointsFuture implements Callable<Object> {

        private final String remoteContainerName;

        public UpdateRemoteNewEndpointsFuture(String remoteContainerName) {
            this.remoteContainerName = remoteContainerName;
        }

        public Object call() throws Exception {

            EndpointRegistryImpl.this.log.debug("Updating remote new endpoints for container : "
                    + this.remoteContainerName);

            try {
                final RemoteJMXServer server = EndpointRegistryImpl.this.jmxService
                        .getRemoteJMXServer(this.remoteContainerName);

                server.invoke(server.getEndpointServiceMBeanName(),
                        RemoteJMXServer.UPDATE_NEW_ENDPOINTS, new Object[] { true },
                        new String[] { boolean.class.getName() });

            } catch (final Exception e) {
                EndpointRegistryImpl.this.log.error(
                        "Failed to update new endpoints on the container '"
                                + this.remoteContainerName + "'", e);
            }

            return null;
        }

    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.registry.EndpointRegistry#activateEndpoint
     * (javax.xml.namespace.QName, java.lang.String,
     * org.ow2.petals.jbi.messaging.endpoint.ConsumerEndpoint)
     */
    public InternalEndpoint activateEndpoint(QName serviceName, String endpointName,
            ConsumerEndpoint address) throws JBIException {
        this.log.start();

        ParameterCheckHelper.isNullParameter(serviceName, "serviceName");
        ParameterCheckHelper.isNullParameter(endpointName, "endpointName");
        ParameterCheckHelper.isNullParameter(address, "address");

        final InternalEndpoint endpoint = new InternalEndpoint(serviceName, endpointName, address
                .getComponentName(), address.getContainerName(), address.getSubDomainName(), this);
        this.registerEndpoint(endpoint);

        // Trigger the timer if it is not yet done
        if (!this.updateScheduled.getAndSet(true)) {
            this.updateLocalNewEndpointsTimer.schedule(new UpdateLocalNewEndpointsTask(), 10000);
        }

        this.log.end();
        return endpoint;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.registry.EndpointRegistry#deactivateEndpoint
     * (javax.jbi.servicedesc.ServiceEndpoint)
     */
    public void deactivateEndpoint(ServiceEndpoint endpoint) throws JBIException {
        this.log.start();

        ParameterCheckHelper.isNullParameter(endpoint, "endpoint");

        final String key = getKey(endpoint);
        try {
            // work either with the retrieved endpoint to be sure that the
            // getters of interfaces/description are functioning
            AbstractEndpoint retrievedEndpoint;

            // Verify that the endpoint is registered
            try {
                retrievedEndpoint = this.retrieveEndpoint(endpoint);
            } catch (final RegistryException e) {
                throw new RegistryException("The endpoint " + endpoint.getEndpointName()
                        + " must be previously registered to be unregistered");
            }

            if (retrievedEndpoint == null) {
                throw new RegistryException("The endpoint " + endpoint.getEndpointName()
                        + " can not be a link to be unregistered");
            }

            // Unbind either the newEndpoint or the registered interfaces
            try {
                this.newEndpointContext.unbind(key);
            } catch (final NameNotFoundException e) {
                for (final QName interfaceName : retrievedEndpoint.getInterfaces()) {
                    final Name itfName = NamingHelper.qNameToName(interfaceName);
                    try {
                        final Context interfaceCtx = (Context) this.interfacesContext
                                .lookup(itfName);
                        interfaceCtx.unbind(key);

                        // delete subcontext if no more endpoint is found
                        if (!this.interfacesContext.list(itfName).hasMore()) {
                            this.interfacesContext.destroySubcontext(itfName);
                        }
                    } catch (final NamingException e1) {
                        throw new RegistryException("Failed to unregister internal endpoint "
                                + retrievedEndpoint.getEndpointName() + ", its interface "
                                + interfaceName + " is not registered", e1);
                    }
                }
            }

            // Unbind from services context
            try {
                final Name serviceName = NamingHelper.qNameToName(retrievedEndpoint
                        .getServiceName());
                final Context serviceCtx = (Context) this.servicesContext.lookup(serviceName);
                serviceCtx.unbind(key);

                // destroy service subcontext if no more endpoints serve
                // this services
                if (!this.servicesContext.list(serviceName).hasMore()) {
                    this.servicesContext.destroySubcontext(serviceName);
                }
            } catch (final NamingException e1) {
                throw new RegistryException("Failed to unregister internal endpoint "
                        + retrievedEndpoint.getEndpointName() + ", its service "
                        + retrievedEndpoint.getServiceName() + " is not registered", e1);
            }

            // Finally unbind the endpoint
            this.endpointsContext.unbind(key);

        } catch (final NamingException e) {
            throw new RegistryException(e);
        }

        this.log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.jbi.messaging.registry.EndpointRegistry#
     * registerExternalEndpoint(javax.jbi.servicedesc.ServiceEndpoint)
     */
    public void registerExternalEndpoint(ServiceEndpoint externalEndpoint) throws JBIException {
        this.log.start();

        if (externalEndpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        this.registerEndpoint(externalEndpoint);

        this.log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.jbi.messaging.registry.EndpointRegistry#
     * deregisterExternalEndpoint(javax.jbi.servicedesc.ServiceEndpoint)
     */
    public void deregisterExternalEndpoint(ServiceEndpoint endpoint) throws JBIException {
        this.log.call();

        this.deactivateEndpoint((AbstractEndpoint) endpoint);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.registry.EndpointRegistry#registerConnection
     * (javax.xml.namespace.QName, javax.xml.namespace.QName, java.lang.String)
     */
    public void registerConnection(QName consInterface, QName provService, String provEndpoint)
            throws JBIException {
        this.log.call();

        ParameterCheckHelper.isNullParameter(consInterface, "consInterface");
        ParameterCheckHelper.isNullParameter(provService, "provService");
        ParameterCheckHelper.isNullParameter(provEndpoint, "provEndpoint");

        final LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consInterface, provService,
                provEndpoint, this);
        this.registerConnection(linkedEndpoint);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.registry.EndpointRegistry#registerConnection
     * (javax.xml.namespace.QName, java.lang.String, javax.xml.namespace.QName,
     * java.lang.String)
     */
    public void registerConnection(QName consService, String consEndpoint, QName provService,
            String provEndpoint) throws JBIException {
        this.log.call();

        ParameterCheckHelper.isNullParameter(consService, "consService");
        ParameterCheckHelper.isNullParameter(consEndpoint, "consEndpoint");
        ParameterCheckHelper.isNullParameter(provService, "provService");
        ParameterCheckHelper.isNullParameter(provEndpoint, "provEndpoint");

        final LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consService, consEndpoint,
                provService, provEndpoint, this);
        this.registerConnection(linkedEndpoint);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.registry.EndpointRegistry#deregisterConnection
     * (javax.xml.namespace.QName, javax.xml.namespace.QName, java.lang.String)
     */
    public void deregisterConnection(QName consInterface, QName provService, String provEndpoint)
            throws JBIException {
        this.log.call();

        ParameterCheckHelper.isNullParameter(consInterface, "consInterface");
        ParameterCheckHelper.isNullParameter(provService, "provService");
        ParameterCheckHelper.isNullParameter(provEndpoint, "provEndpoint");

        final LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consInterface, provService,
                provEndpoint, this);
        this.unregisterConnection(linkedEndpoint);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.registry.EndpointRegistry#deregisterConnection
     * (javax.xml.namespace.QName, java.lang.String, javax.xml.namespace.QName,
     * java.lang.String)
     */
    public void deregisterConnection(QName consService, String consEndpoint, QName provService,
            String provEndpoint) throws JBIException {
        this.log.call();

        ParameterCheckHelper.isNullParameter(consService, "consService");
        ParameterCheckHelper.isNullParameter(consEndpoint, "consEndpoint");
        ParameterCheckHelper.isNullParameter(provService, "provService");
        ParameterCheckHelper.isNullParameter(provEndpoint, "provEndpoint");

        final LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consService, consEndpoint,
                provService, provEndpoint, this);
        this.unregisterConnection(linkedEndpoint);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.registry.EndpointRegistry#getEndpoint(javax
     * .xml.namespace.QName, java.lang.String)
     */
    public ServiceEndpoint getEndpoint(QName service, String name) {
        this.log.call();

        AbstractEndpoint abstractEndpoint = null;
        try {
            // TODO: retrieve external endpoint too?
            abstractEndpoint = this.retrieveEndpoint(service, name, InternalEndpoint.class
                    .getName(), LinkType.STANDARD, false);
            if (abstractEndpoint != null) {
                if (!this.isVisibleEndpoint(abstractEndpoint)) {
                    abstractEndpoint = null;
                }
            }
        } catch (final RegistryException e) {
            // no endpoint found, return null;
        }

        return abstractEndpoint;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.jbi.messaging.registry.EndpointRegistry#
     * getEndpointDescriptorForEndpoint(javax.jbi.servicedesc.ServiceEndpoint)
     */
    public Document getEndpointDescriptorForEndpoint(ServiceEndpoint endpoint) throws JBIException {
        this.log.start();

        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        Document document = null;

        if ((endpoint instanceof AbstractEndpoint)) {
            document = ((AbstractEndpoint) endpoint).getDescription();
            if (document == null) {
                try {
                    // update the new endpoints if necessary
                    if (this.isNewEndpoint(endpoint)) {
                        synchronized (this.containerName) {
                            this.updateNewEndpoints(false);
                        }
                    }

                    // retrieve the endpoint
                    final AbstractEndpoint abstractEndpoint = this.retrieveEndpoint(endpoint);
                    document = abstractEndpoint.getDescription();

                    // set both description and interfaces
                    ((AbstractEndpoint) endpoint).setDescription(document);
                    ((AbstractEndpoint) endpoint).setInterfaces(abstractEndpoint.getInterfaces());
                } catch (final RegistryException e) {
                    this.log.error("Error while validating endpoint " + endpoint.getEndpointName()
                            + " for service " + endpoint.getServiceName(), e);
                }
            }
        }

        this.log.end();
        return document;
    }

    /**
     * Return the interface for the specified endpoint which is not a
     * connection.
     * 
     * @param endpoint
     *            endpoint on which we want the interface names; must be non
     *            null
     * @return array of interfaces QName
     */
    public QName[] getInterfacesForEndpoint(AbstractEndpoint endpoint) {
        this.log.start();

        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }

        // update only the first time
        if (endpoint.getInterfaces() == null) {
            try {
                if (this.isNewEndpoint(endpoint)) {
                    synchronized (this.containerName) {
                        this.updateNewEndpoints(false);
                    }
                }

                // retrieve the endpoint
                final AbstractEndpoint abstractEndpoint = this.retrieveEndpoint(endpoint);

                // set both description and interfaces
                endpoint.setInterfaces(abstractEndpoint.getInterfaces());
                endpoint.setDescription(abstractEndpoint.getDescription());
            } catch (final Exception e) {
                this.log.error("Error while getting interfaces for endpoint "
                        + endpoint.getEndpointName() + " of service " + endpoint.getServiceName(),
                        e);
            }
        }

        this.log.end();
        return endpoint.getInterfaces();
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.jbi.messaging.registry.EndpointRegistry#
     * getInternalEndpointsForInterface(javax.xml.namespace.QName,
     * org.ow2.petals.jbi.descriptor.original.generated.LinkType)
     */
    public ServiceEndpoint[] getInternalEndpointsForInterface(QName interfaceName, LinkType linkType) {
        this.log.start();

        ServiceEndpoint[] endpoints = new ServiceEndpoint[0];
        try {
            // Update the new endpoints
            synchronized (this.containerName) {
                this.updateNewEndpoints(false);
            }

            endpoints = this.removeNonVisibleEndpoints(this.retrieveEndpointsForInterface(
                    interfaceName, InternalEndpoint.class.getName(), linkType));

        } catch (final NamingException e) {
            this.log.error("Error while getting internal endpoints for interface " + interfaceName,
                    e);
        } catch (final JBIException e) {
            this.log.error("Error while getting internal endpoints for interface " + interfaceName,
                    e);
        }

        this.log.end();
        return endpoints;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.jbi.messaging.registry.EndpointRegistry#
     * getInternalEndpointsForService(javax.xml.namespace.QName,
     * org.ow2.petals.jbi.descriptor.original.generated.LinkType)
     */
    public ServiceEndpoint[] getInternalEndpointsForService(QName serviceName, LinkType linkType) {
        this.log.start();

        if (serviceName == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }

        ServiceEndpoint[] endpoints = new ServiceEndpoint[0];
        try {
            endpoints = this.removeNonVisibleEndpoints(this.retrieveEndpointsForService(
                    serviceName, InternalEndpoint.class.getName(), linkType));
        } catch (final NamingException e) {
            this.log.error("Error while getting internal endpoints for service " + serviceName, e);
        }

        this.log.end();
        return endpoints;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.jbi.messaging.registry.EndpointRegistry#
     * getExternalEndpointsForInterface(javax.xml.namespace.QName)
     */
    public ServiceEndpoint[] getExternalEndpointsForInterface(QName interfaceName) {
        this.log.start();

        ServiceEndpoint[] endpoints = new ServiceEndpoint[0];
        try {
            endpoints = this.removeNonVisibleEndpoints(this.retrieveEndpointsForInterface(
                    interfaceName, ExternalEndpoint.class.getName(), LinkType.STANDARD));
        } catch (final NamingException e) {
            this.log.error("Error while getting external endpoints for interface " + interfaceName,
                    e);
        }
        this.log.end();
        return endpoints;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.jbi.messaging.registry.EndpointRegistry#
     * getExternalEndpointsForService(javax.xml.namespace.QName)
     */
    public ServiceEndpoint[] getExternalEndpointsForService(QName serviceName) {
        this.log.start();

        if (serviceName == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        ServiceEndpoint[] endpoints = new ServiceEndpoint[0];
        try {
            endpoints = this.removeNonVisibleEndpoints(this.retrieveEndpointsForService(
                    serviceName, ExternalEndpoint.class.getName(), LinkType.STANDARD));
        } catch (final NamingException e) {
            this.log.error("Error while getting external endpoints for service " + serviceName, e);
        }

        this.log.end();
        return endpoints;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.jbi.messaging.registry.EndpointRegistry#
     * isExchangeWithConsumerOkayForComponent
     * (org.ow2.petals.jbi.messaging.endpoint.InternalEndpoint,
     * javax.jbi.messaging.MessageExchange)
     */
    public boolean isExchangeWithConsumerOkayForComponent(InternalEndpoint internalEndpoint,
            MessageExchange exchange) {
        this.log.start();
        ParameterCheckHelper.isNullParameterWithLog(internalEndpoint, "Endpoint", this.log);
        boolean test = true;
        try {
            // Retrieve a connection to the JMX server of the container were the
            // component is deployed
            final RemoteJMXServer server = this.jmxService.getRemoteJMXServer(internalEndpoint
                    .getContainerName());
            // Ask the component if it is ok to accept the exchange
            final Boolean bool = (Boolean) server.invoke(server.getAdminServiceMBeanName(),
                    RemoteJMXServer.IS_OK_WITH_CONS, new Object[] {
                            internalEndpoint.getComponentName(), internalEndpoint, exchange },
                    new String[] { String.class.getName(), ServiceEndpoint.class.getName(),
                            MessageExchange.class.getName() });
            test = bool.booleanValue();
        } catch (final Exception e) {
            this.log.error("Problem in isExchangeWithConsumerOkayForComponent", e);
        }
        this.log.end();
        return test;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.jbi.messaging.registry.EndpointRegistry#
     * isExchangeWithProviderOkayForComponent
     * (org.ow2.petals.jbi.messaging.endpoint.InternalEndpoint,
     * javax.jbi.messaging.MessageExchange)
     */
    public boolean isExchangeWithProviderOkayForComponent(InternalEndpoint internalEndpoint,
            MessageExchange exchange) {
        this.log.start();
        if (internalEndpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        boolean test = true;
        try {
            // Retrieve a connection to the JMX server of the container were the
            // component is deployed
            final RemoteJMXServer server = this.jmxService.getRemoteJMXServer(internalEndpoint
                    .getContainerName());
            // Retrieve the endpoint description
            final Boolean bool = (Boolean) server.invoke(server.getAdminServiceMBeanName(),
                    RemoteJMXServer.IS_OK_WITH_PROV, new Object[] {
                            internalEndpoint.getComponentName(), internalEndpoint, exchange },
                    new String[] { String.class.getName(), ServiceEndpoint.class.getName(),
                            MessageExchange.class.getName() });
            test = bool.booleanValue();
        } catch (final Exception e) {
            this.log.error("Problem in isExchangeWithProviderOkayForComponent", e);
        }
        this.log.end();
        return test;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.registry.EndpointRegistryMBean#getDescription
     * (java.lang.String, java.lang.String)
     */
    public Document getDescription(String serviceName, String endpointName) throws JBIException {
        this.log.call();

        Document resultDocument = null;

        final ServiceEndpoint serviceEndpoint = this.getEndpoint(QName.valueOf(serviceName),
                endpointName);

        if (serviceEndpoint != null) {
            resultDocument = this.getEndpointDescriptorForEndpoint(serviceEndpoint);
        }

        return resultDocument;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.jbi.messaging.registry.EndpointRegistryMBean#
     * getAllExternalEndpoints()
     */
    public List<Map<String, Object>> getAllExternalEndpoints() {
        this.log.start();

        final List<Map<String, Object>> externalEndpointsMap = new ArrayList<Map<String, Object>>();
        final ServiceEndpoint[] externalEndpoints = this.getExternalEndpointsForInterface(null);

        for (final ServiceEndpoint serviceEndpoint : externalEndpoints) {
            final AbstractEndpoint abstractEndpoint = (AbstractEndpoint) serviceEndpoint;
            final Map<String, Object> endpointMap = new HashMap<String, Object>();
            endpointMap.put(KEY_COMPONENT_NAME, abstractEndpoint.getComponentName());
            endpointMap.put(KEY_CONTAINER_NAME, abstractEndpoint.getContainerName());
            endpointMap.put(KEY_ENDPOINT_NAME, abstractEndpoint.getEndpointName());
            endpointMap.put(KEY_SERVICE_NAME, abstractEndpoint.getServiceName());
            final List<String> interfaceNames = interfaceQNameToString(abstractEndpoint);
            endpointMap.put(KEY_INTERFACE_NAMES, interfaceNames.toArray(new String[0]));
            endpointMap.put(KEY_TYPE, TYPE_EXTERNAL);
            externalEndpointsMap.add(endpointMap);
        }

        this.log.end();
        return externalEndpointsMap;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.jbi.messaging.registry.EndpointRegistryMBean#
     * getAllInternalEndpoints()
     */
    public List<Map<String, Object>> getAllInternalEndpoints() {
        this.log.start();

        final List<Map<String, Object>> internalEndpointsMap = new ArrayList<Map<String, Object>>();
        final ServiceEndpoint[] internalEndpoints = this.getInternalEndpointsForInterface(null,
                LinkType.STANDARD);

        for (final ServiceEndpoint serviceEndpoint : internalEndpoints) {
            final AbstractEndpoint abstractEndpoint = (AbstractEndpoint) serviceEndpoint;
            final Map<String, Object> endpointMap = new HashMap<String, Object>();
            endpointMap.put(KEY_COMPONENT_NAME, abstractEndpoint.getComponentName());
            endpointMap.put(KEY_CONTAINER_NAME, abstractEndpoint.getContainerName());
            endpointMap.put(KEY_ENDPOINT_NAME, abstractEndpoint.getEndpointName());
            endpointMap.put(KEY_SERVICE_NAME, abstractEndpoint.getServiceName().toString());
            final List<String> interfaceNames = interfaceQNameToString(abstractEndpoint);
            endpointMap.put(KEY_INTERFACE_NAMES, interfaceNames.toArray(new String[0]));
            if (serviceEndpoint instanceof LinkedEndpoint) {
                final LinkedEndpoint linkEP = (LinkedEndpoint) serviceEndpoint;
                endpointMap.put(KEY_TYPE, linkEP.getToServiceName().toString() + "@"
                        + linkEP.getToEndpointName());
            } else {
                endpointMap.put(KEY_TYPE, TYPE_INTERNAL);
            }
            internalEndpointsMap.add(endpointMap);
        }

        this.log.end();
        return internalEndpointsMap;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.registry.EndpointRegistryMBean#getAllEndpoints
     * ()
     */
    public List<Map<String, Object>> getAllEndpoints() {
        this.log.call();

        final List<Map<String, Object>> allEndpointsMap = this.getAllExternalEndpoints();
        allEndpointsMap.addAll(this.getAllInternalEndpoints());

        return allEndpointsMap;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.jbi.messaging.registry.EndpointRegistryMBean#
     * updateNewEndpoints(boolean)
     */
    public void updateNewEndpoints(boolean remoteContainerInvocation) throws JBIException {
        this.log.start("RemoteContainerInvocation : " + remoteContainerInvocation);

        final List<AbstractEndpoint> newLocalEndpoints = new ArrayList<AbstractEndpoint>();
        final List<String> containersToUpdate = new ArrayList<String>();
        List<Future<Object>> futures;

        // Synchronize this part if a remote and a local invocation are running
        // concurrently
        synchronized (this.containerName) {
            this.retrieveNewEndpoints(newLocalEndpoints, containersToUpdate);
            futures = new ArrayList<Future<Object>>(newLocalEndpoints.size());

            // Update the local new endpoints asynchronously
            for (final AbstractEndpoint abstractEndpoint : newLocalEndpoints) {
                futures.add(this.updateNewEndpointsCS.submit(new UpdateLocalNewEndpointFuture(
                        abstractEndpoint)));
            }

            for (final Future<Object> future : futures) {
                try {
                    future.get();
                } catch (final InterruptedException e) {
                    this.log.warning("Update of a new endpoint has been interrupted");
                } catch (final ExecutionException e) {
                    this.log.warning("An error occurs during an update of a new endpoint", e
                            .getCause());
                }
            }
        }

        // Invoke update of remote containers only once
        if (!remoteContainerInvocation) {
            futures = new ArrayList<Future<Object>>(containersToUpdate.size());

            // Trigger remote update asynchronously
            for (final String containerToUpdate : containersToUpdate) {
                this.updateNewEndpointsCS.submit(new UpdateRemoteNewEndpointsFuture(
                        containerToUpdate));
            }
            for (final Future<Object> future : futures) {
                try {
                    future.get();
                } catch (final InterruptedException e) {
                    this.log.warning("Update of a new endpoint has been interrupted");
                } catch (final ExecutionException e) {
                    this.log.warning("An error occurs during an update of a new endpoint", e
                            .getCause());
                }
            }
        }

        this.log.end();
    }

    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws RegistryException, WSDLException {
        this.log = new LoggingUtil(this.logger);
        this.log.start();

        try {
            this.wsdlReader = WSDL4ComplexWsdlFactory.newInstance().newWSDLReader();
        } catch (final WSDLException e) {
            this.log.error("Failed to start Endpoint Registry service", e);
            throw e;
        }

        this.containerName = this.configurationService.getContainerConfiguration().getName();
        this.subdomainName = this.configurationService.getSubDomainConfiguration().getName();
        this.updateLocalNewEndpointsTimer = new Timer(true);
        this.updateScheduled = new AtomicBoolean(false);
        this.updateNewEndpointsTPE = new ThreadPoolExecutor(1, 50, 60, TimeUnit.SECONDS,
                new LinkedBlockingQueue<Runnable>());
        this.updateNewEndpointsCS = new ExecutorCompletionService<Object>(
                this.updateNewEndpointsTPE);

        try {
            this.initContexts();
        } catch (final RegistryException e) {
            this.log.error("Failed to start Endpoint Registry service", e);
            throw e;
        }

        this.log.end();
    }

    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() throws IllegalLifeCycleException {
        this.log.call();

        this.updateLocalNewEndpointsTimer.cancel();
        this.updateNewEndpointsTPE.shutdownNow();
    }

    /**
     * Create and / or reference contexts "containers", "endpoints", "services"
     * and "interfaces" in the global directory.
     * 
     * @throws JBIException
     *             problems while creating or lookup contexts
     */
    private void initContexts() throws RegistryException {
        this.log.start();
        try {
            this.rootContext = this.jndiService.getInitialContext();

            try {
                this.endpointsContext = (Context) this.rootContext.lookup(ENDPOINTS_REF);
            } catch (final NameNotFoundException e) {
                this.endpointsContext = this.rootContext.createSubcontext(ENDPOINTS_REF);
            }

            try {
                this.servicesContext = (Context) this.rootContext.lookup(SERVICES_REF);
            } catch (final NameNotFoundException e) {
                this.servicesContext = this.rootContext.createSubcontext(SERVICES_REF);
            }

            try {
                this.newEndpointContext = (Context) this.rootContext.lookup(NEW_ENDPOINTS_REF);
            } catch (final NameNotFoundException e) {
                this.newEndpointContext = this.rootContext.createSubcontext(NEW_ENDPOINTS_REF);
            }

            try {
                this.interfacesContext = (Context) this.rootContext.lookup(INTERFACES_REF);
            } catch (final NameNotFoundException e) {
                this.interfacesContext = this.rootContext.createSubcontext(INTERFACES_REF);
            }

            if (this.endpointsContext == null || this.servicesContext == null
                    || this.interfacesContext == null) {
                throw new NotContextException();
            }
        } catch (final Exception e) {
            throw new RegistryException(e);
        }
        this.log.end();
    }

    /**
     * Unregister a connection
     * 
     * @param connection
     *            connection to unregistered; must be non null
     * @throws RegistryException
     *             if the connection was not previously registered
     */
    private void unregisterConnection(LinkedEndpoint linkedEndpoint) throws RegistryException {
        this.log.start();

        final String toKey = getKey(linkedEndpoint.getToServiceName(), linkedEndpoint
                .getToEndpointName());

        final String key = getKey(linkedEndpoint.getServiceName(), linkedEndpoint.getEndpointName());

        if (linkedEndpoint.getInterfaces() != null) {
            final Name interf = NamingHelper.qNameToName(linkedEndpoint.getInterfaces()[0]);
            try {
                try {
                    // check if the provider endpoint is still bound in the
                    // new endpoint context if so its interfaces have not
                    // been resolved
                    this.newEndpointContext.lookup(toKey);
                    final Context interfaceCtx = (Context) this.interfacesContext.lookup(interf);
                    interfaceCtx.unbind(key);
                } catch (final NameNotFoundException e) {
                    // we need to check that we do not remove it from
                    // one of its interfaces context
                    final InternalEndpoint internalEndpoint = (InternalEndpoint) this
                            .retrieveEndpoint(linkedEndpoint.getToServiceName(), linkedEndpoint
                                    .getEndpointName(), InternalEndpoint.class.getName(),
                                    LinkType.SOFT, true);
                    if (internalEndpoint == null) {
                        throw new NameNotFoundException(linkedEndpoint.getEndpointName()
                                + " not found");
                    }
                    boolean isNotInInt = true;
                    for (final QName i : internalEndpoint.getInterfaces()) {
                        if (linkedEndpoint.getInterfaces()[0].equals(i)) {
                            isNotInInt = false;
                        }
                    }
                    if (isNotInInt) {
                        final Context interfaceCtx = (Context) this.interfacesContext
                                .lookup(interf);
                        interfaceCtx.unbind(key);
                        if (!this.interfacesContext.list(interf).hasMoreElements()) {
                            this.interfacesContext.destroySubcontext(interf);
                        }
                    }
                }
            } catch (final NamingException e) {
                throw new RegistryException("Failed to unregister connection with interface name "
                        + linkedEndpoint.getInterfaces()[0] + " to "
                        + linkedEndpoint.getToEndpointName(), e);
            }
        } else {
            try {
                // If the provider service name is different from the consumer
                // service
                // name we removed it
                if (!linkedEndpoint.getToServiceName().equals(linkedEndpoint.getServiceName())) {
                    this.retrieveServiceContext(
                            NamingHelper.qNameToName(linkedEndpoint.getServiceName())).unbind(key);
                    // If the service context is empty, we need to remove it
                    final Name serviceName = NamingHelper.qNameToName(linkedEndpoint
                            .getServiceName());
                    if (!this.servicesContext.list(serviceName).hasMore()) {
                        this.servicesContext.destroySubcontext(serviceName);
                    }
                }

                // If the provider endpoint name is different from the consumer
                // endpoint
                // name we removed it
                // FIXME : Test the service name too ???
                if (!toKey.equals(key)) {
                    this.endpointsContext.unbind(key);
                }

            } catch (final NamingException e) {
                throw new RegistryException("Failed to unregister connection with service name "
                        + linkedEndpoint.getServiceName() + " and endpoint name "
                        + linkedEndpoint.getEndpointName() + " to "
                        + linkedEndpoint.getToEndpointName(), e);
            }
        }

        this.log.end();
    }

    private void registerEndpoint(ServiceEndpoint endpoint) throws RegistryException {
        this.log.start();

        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must not be null");
        }

        AbstractEndpoint registeredEndpoint = null;
        boolean found = false;

        try {
            // Retrieve a potential matching endpoint
            registeredEndpoint = this.retrieveEndpoint(endpoint);
            found = true;
        } catch (final RegistryException e) {
            final String key = getKey(endpoint);

            try {
                // register endpoint
                this.endpointsContext.bind(key, endpoint);

                // register newEndpoint
                final LinkRef linkRef = new LinkRef("/" + ENDPOINTS_REF + "/" + key);
                this.newEndpointContext.bind(key, linkRef);

                // register service
                final Name serviceName = NamingHelper.qNameToName(endpoint.getServiceName());
                if (serviceName != null) {
                    final Context servContext = this.retrieveServiceContext(serviceName);
                    try {
                        servContext.bind(key, linkRef);
                    } catch (final NameAlreadyBoundException e2) {
                        // do nothing
                    }
                }
            } catch (final NamingException e1) {
                throw new RegistryException("Problem while registering endpoint "
                        + endpoint.getEndpointName(), e1);
            }
        }

        if (found) {
            // registeredEndpoint == null ==> the bound object isn't a real
            // endpoint (link or anything else)
            if (registeredEndpoint == null) {
                throw new RegistryException(
                        "You are trying to register an endpoint with a name that "
                                + "is already bound to an other object "
                                + "(link or anything else) in the JNDI Registry.");
            }

            // An endpoint can't be registered with the same endpoint name as a
            // previously registered endpoint that have different container name
            // or component name
            if (!(((AbstractEndpoint) endpoint).getContainerName().equals(
                    registeredEndpoint.getContainerName()) && ((AbstractEndpoint) endpoint)
                    .getComponentName().equals(registeredEndpoint.getComponentName()))) {
                throw new RegistryException("The endpoint " + endpoint.getEndpointName()
                        + " is already registered for an other component/container");
            }
        }

        this.log.end();
    }

    /**
     * Register a connection
     * 
     * @param connection
     *            connection to registered; must be non null
     * @throws RegistryException
     *             if the connection has already been registered or if a problem
     *             occurs
     */
    private void registerConnection(LinkedEndpoint linkedEndpoint) throws RegistryException {
        this.log.start();

        // targeted endpoint
        final String toKey = getKey(linkedEndpoint.getToServiceName(), linkedEndpoint
                .getToEndpointName());

        // ' virtual' endpoint
        final String key = getKey(linkedEndpoint.getServiceName(), linkedEndpoint.getEndpointName());

        final LinkRef linkRef = new LinkRef("/" + ENDPOINTS_REF + "/" + toKey);

        if (linkedEndpoint.getInterfaces() != null) {
            try {
                final Name itfName = NamingHelper.qNameToName(linkedEndpoint.getInterfaces()[0]);

                // bind or rebind link in interfaces
                final Context itfContext = this.retrieveInterfaceContext(itfName);
                itfContext.rebind(key, linkRef);

            } catch (final NamingException e) {
                throw new RegistryException(
                        "Problem while registering connection with interfacename "
                                + linkedEndpoint.getInterfaces()[0] + " to "
                                + linkedEndpoint.getToEndpointName() + " ");
            }
        } else {
            try {
                final Name serviceName = NamingHelper.qNameToName(linkedEndpoint.getServiceName());

                // bind or rebind service
                final Context servContext = this.retrieveServiceContext(serviceName);
                servContext.rebind(key, linkRef);

                // bind or rebind endpoint
                this.endpointsContext.rebind(key, linkRef);

            } catch (final NamingException e) {
                throw new RegistryException(
                        "Problem while registering connection with service name  "
                                + linkedEndpoint.getServiceName() + " and endpoint name "
                                + linkedEndpoint.getEndpointName() + " to "
                                + linkedEndpoint.getToEndpointName() + " ");
            }
        }

        this.log.end();
    }

    /**
     * Retrieve the list of all the new endpoints. This method is synchronized
     * to avoid concurrent invocations.
     * 
     * @param localNewEndpoints
     *            The list of the local new endpoints to fill
     * @param containerToUpdate
     *            The list of containerToUpdate to fill
     * 
     * @throws RegistryException
     */
    private synchronized void retrieveNewEndpoints(List<AbstractEndpoint> localNewEndpoints,
            List<String> containerToUpdate) throws RegistryException {
        this.log.start();

        try {
            final NamingEnumeration<Binding> bindings = this.rootContext
                    .listBindings(NEW_ENDPOINTS_REF);

            while (bindings.hasMoreElements()) {
                final Binding binding = bindings.next();
                final Object object = binding.getObject();

                if (object instanceof LinkRef) {
                    final String newEndpointName = ((LinkRef) object).getLinkName();
                    try {
                        final AbstractEndpoint abstractEndpoint = (AbstractEndpoint) this.rootContext
                                .lookup(newEndpointName);
                        if (this.containerName.equals(abstractEndpoint.getContainerName())) {
                            localNewEndpoints.add(abstractEndpoint);
                        } else if (!containerToUpdate.contains(abstractEndpoint.getContainerName())) {
                            containerToUpdate.add(abstractEndpoint.getContainerName());
                        }

                    } catch (final NameNotFoundException e1) {
                        // The new endpoint may be updated concurrently by
                        // another node, do nothing
                    }
                } else {
                    this.log.error("The new endpoint entry named '" + binding.getName()
                            + "' must be a " + LinkRef.class + " instance. It will be deleted");
                    this.newEndpointContext.unbind(binding.getName());
                }
            }

        } catch (final NamingException e) {
            throw new RegistryException("Problem while retrieving new endpoints", e);
        }

        this.log.end();
    }

    /**
     * Update the endpoint, remove it from the new-endpoint context and update
     * the interfaces context
     * 
     * @param endpoint
     *            endpoint to validate; must be non null
     * @throws RegistryException
     *             if a problem occurs
     */
    private void validateEndpoint(final AbstractEndpoint endpoint) throws RegistryException {
        this.log.start();

        final String key = getKey(endpoint);

        try {
            // Add the interfaces to the interfaces context
            for (final QName name : endpoint.getInterfaces()) {
                final Name itfName = NamingHelper.qNameToName(name);
                final LinkRef linkRef = new LinkRef("/" + ENDPOINTS_REF + "/" + key);

                // bind or rebind endpoint in interfaces
                final Context interfaceCtx = this.retrieveInterfaceContext(itfName);
                interfaceCtx.rebind(key, linkRef);
            }

            // Update the endpoint in the endpoint context
            this.endpointsContext.rebind(key, endpoint);

        } catch (final NamingException e) {
            throw new RegistryException(e);
        } finally {
            try {
                this.newEndpointContext.unbind(key);
            } catch (final NamingException e) {
                // do nothing
            }
        }

        this.log.end();
    }

    /**
     * Get the internal/external service endpoint for the named activated
     * endpoint, if any
     * 
     * @param service
     *            qualified service name
     * @param endpointName
     *            endpoint name
     * @param endpointClassName
     *            class name of the endpoint
     * @param resolveLink
     *            if the Endpoint is a link, resolve it (true) or return the
     *            link (false)
     * @return
     * @throws RegistryException
     */
    private AbstractEndpoint retrieveEndpoint(QName serviceName, String endpointName,
            String endpointClassName, LinkType linkType, boolean resolveLink)
            throws RegistryException {
        this.log.start();

        AbstractEndpoint result = null;
        final String key = getKey(serviceName, endpointName);

        try {

            final Object outObj = this.endpointsContext.lookup(key);

            // real endpoint
            if (outObj instanceof AbstractEndpoint
                    && (linkType == LinkType.STANDARD || linkType == LinkType.HARD)) {
                final AbstractEndpoint endpoint = (AbstractEndpoint) outObj;
                if (endpoint.getClass().getName().equals(endpointClassName)) {
                    endpoint.setEndpointRegistry(this);
                    result = endpoint;
                }
            }
            // link endpoint
            else if (outObj instanceof LinkRef
                    && (linkType == LinkType.STANDARD || linkType == LinkType.SOFT)) {
                final Object obj = this.rootContext.lookup(((LinkRef) outObj).getLinkName());
                if (obj instanceof AbstractEndpoint) {
                    final AbstractEndpoint endpoint = (AbstractEndpoint) obj;
                    if (endpoint.getClass().getName().equals(endpointClassName)) {
                        // return the real endpoint
                        if (resolveLink) {
                            endpoint.setEndpointRegistry(this);
                            result = endpoint;
                        }
                        // return a link
                        else {
                            result = new LinkedEndpoint(serviceName, endpointName, endpoint
                                    .getServiceName(), endpoint.getEndpointName(), null);
                        }
                    }
                }
            }
        } catch (final NamingException e1) {
            // Do nothing
        }

        if (result != null) {
            if (!this.isVisibleEndpoint(result)) {
                result = null;
            }
        }

        this.log.end();
        return result;
    }

    /**
     * Retrieve an endpoint in the JNDI registry without making distinction
     * between Internal and External one.
     * 
     * @param endpoint
     *            endpoint description
     * @return an {@link AbstractEndpoint} if the matching endpoint is a real
     *         endpoint, null otherwise (for exemple, the matching endpoint is a
     *         link)
     * 
     * @throws RegistryException
     */
    private AbstractEndpoint retrieveEndpoint(ServiceEndpoint endpoint) throws RegistryException {
        this.log.start();

        AbstractEndpoint result = null;
        try {
            final Object outObj = this.endpointsContext.lookup(getKey(endpoint));

            // real endpoint
            if (outObj instanceof AbstractEndpoint) {
                result = (AbstractEndpoint) outObj;
                result.setEndpointRegistry(this);
            }
        } catch (final NamingException e) {
            throw new RegistryException("Error while looking up an Endpoint in the JNDI Registry",
                    e);
        }
        this.log.end();
        return result;
    }

    /**
     * Retrieve the list of endpoint that implements the service name. If
     * connections are included they are resolved.
     * 
     * @param serviceName
     * @param endpointType
     * @return
     * @throws NamingException
     */
    private AbstractEndpoint[] retrieveEndpointsForService(QName serviceName, String endpointType,
            LinkType linkType) throws NamingException {
        this.log.start();

        final Name srvName = NamingHelper.qNameToName(serviceName);
        final NamingEnumeration<Binding> bindings = this.servicesContext.listBindings(srvName);
        final List<AbstractEndpoint> endpoints = new ArrayList<AbstractEndpoint>();

        while (bindings.hasMoreElements()) {
            final Binding binding = bindings.next();
            final Object object = binding.getObject();

            // real endpoint case
            if (object instanceof AbstractEndpoint
                    && (linkType == LinkType.STANDARD || linkType == LinkType.HARD)) {
                final AbstractEndpoint endpoint = (AbstractEndpoint) object;
                if (endpoint.getClass().getName().equals(endpointType)) {
                    endpoint.setEndpointRegistry(this);
                    endpoints.add(endpoint);
                }
            }

            // connection endpoint case
            if (object instanceof LinkRef
                    && (linkType == LinkType.STANDARD || linkType == LinkType.SOFT)) {
                final LinkRef linkRef = (LinkRef) object;
                final Object obj = this.rootContext.lookup(linkRef.getLinkName());
                if (obj instanceof AbstractEndpoint) {
                    final AbstractEndpoint endpoint = (AbstractEndpoint) obj;
                    if (endpoint.getClass().getName().equals(endpointType)) {
                        endpoint.setEndpointRegistry(this);
                        endpoints.add(endpoint);
                    }
                }
            }
        }

        this.log.end();
        return endpoints.toArray(new AbstractEndpoint[0]);
    }

    /**
     * Retrieve the list of endpoint that implements the interface name. If
     * connections are included they are resolved.
     * 
     * @param interfaceName
     *            the {@link QName} of the interface
     * @param endpointType
     *            the type (class name) of the endpoint (InternalEndpoint or
     *            ExternalEndpoint)
     * @return an Array of matching endpoints
     * @throws RegistryException
     */
    private AbstractEndpoint[] retrieveEndpointsForInterface(QName interfaceName,
            String endpointType, LinkType linkType) throws NamingException {
        this.log.start();

        final List<AbstractEndpoint> endpoints = new ArrayList<AbstractEndpoint>();
        if (interfaceName == null) {
            // We must return all the endpoints of the specified type
            final NamingEnumeration<Binding> bindings = this.rootContext
                    .listBindings(ENDPOINTS_REF);

            while (bindings.hasMoreElements()) {
                final Binding binding = bindings.next();
                final Object object = binding.getObject();

                // real endpoint case
                if (object instanceof AbstractEndpoint
                        && (linkType == LinkType.STANDARD || linkType == LinkType.HARD)) {
                    final AbstractEndpoint endpoint = (AbstractEndpoint) object;
                    if (endpoint.getClass().getName().equals(endpointType)) {
                        endpoint.setEndpointRegistry(this);
                        endpoints.add(endpoint);
                    }
                }

                // connection endpoint case
                else if (object instanceof LinkRef
                        && (linkType == LinkType.STANDARD || linkType == LinkType.SOFT)) {
                    final LinkRef linkRef = (LinkRef) object;
                    final Object obj = this.rootContext.lookup(linkRef.getLinkName());
                    // connection linked to a real EP
                    if (obj instanceof AbstractEndpoint) {
                        final AbstractEndpoint endpoint = (AbstractEndpoint) obj;
                        if (endpoint.getClass().getName().equals(endpointType)) {
                            final LinkedEndpoint epLink = new LinkedEndpoint(
                                    getServiceNameFromKey(binding.getName()),
                                    getEndpointNameFromKey(binding.getName()), endpoint
                                            .getServiceName(), endpoint.getEndpointName(), null);
                            endpoints.add(epLink);
                        }
                    }
                }
            }
        } else {
            // We must return only the endpoints for the specified
            // interface
            final Name itfName = NamingHelper.qNameToName(interfaceName);
            final NamingEnumeration<Binding> bindings = this.interfacesContext
                    .listBindings(itfName);

            while (bindings.hasMoreElements()) {
                final Binding binding = bindings.next();
                final Object object = binding.getObject();

                // real endpoint case
                if (object instanceof AbstractEndpoint
                        && (linkType == LinkType.STANDARD || linkType == LinkType.HARD)) {
                    final AbstractEndpoint endpoint = (AbstractEndpoint) object;
                    if (endpoint.getClass().getName().equals(endpointType)) {
                        endpoint.setEndpointRegistry(this);
                        endpoints.add(endpoint);
                    }
                }

                // connection endpoint case
                if (object instanceof LinkRef) {
                    final LinkRef linkRef = (LinkRef) object;
                    final Object obj = this.rootContext.lookup(linkRef.getLinkName());
                    if (obj instanceof AbstractEndpoint) {
                        final AbstractEndpoint endpoint = (AbstractEndpoint) obj;
                        if (endpoint.getClass().getName().equals(endpointType)) {
                            endpoint.setEndpointRegistry(this);
                            endpoints.add(endpoint);
                        }
                    }
                }
            }
        }

        this.log.end();
        return endpoints.toArray(new AbstractEndpoint[0]);
    }

    /**
     * Return the interfaceContext associated with this interfaceName. If the
     * subcontext does not exists, it is created.
     * 
     * @param serviceName
     * @throws NamingException
     */
    private Context retrieveInterfaceContext(Name interfaceName) throws NamingException {
        this.log.call();

        Context interfContext = null;
        try {
            interfContext = (Context) this.interfacesContext.lookup(interfaceName);
        } catch (final NameNotFoundException e) {
            interfContext = this.interfacesContext.createSubcontext(interfaceName);
        }
        return interfContext;
    }

    /**
     * Return the serviceContext associated with this serviceName. If the
     * subcontext does not exists, it is created.
     * 
     * @param serviceName
     * @throws NamingException
     */
    private Context retrieveServiceContext(Name serviceName) throws NamingException {
        this.log.call();

        Context serviceContext = null;

        try {
            serviceContext = (Context) this.servicesContext.lookup(serviceName);
        } catch (final NameNotFoundException e) {
            serviceContext = this.servicesContext.createSubcontext(serviceName);
        }

        return serviceContext;
    }

    /**
     * Remove from the collection the endpoints that must not be shown, ie
     * private endpoints from other sub-domains
     * 
     * @return a collection of visible endpoints
     */
    private AbstractEndpoint[] removeNonVisibleEndpoints(AbstractEndpoint[] endpoints) {
        final ArrayList<AbstractEndpoint> result = new ArrayList<AbstractEndpoint>();
        for (final AbstractEndpoint abstractEndpoint : endpoints) {
            if (this.isVisibleEndpoint(abstractEndpoint)) {
                result.add(abstractEndpoint);
            }
        }
        return result.toArray(new AbstractEndpoint[0]);
    }

    /**
     * return true if the endpoint is Visible <br>
     * (same sub domain as the local container or public endpoint) <br>
     * public services does not contains "private" in the namespace
     * 
     * @param abstractEndpoint
     * @return
     */
    private boolean isVisibleEndpoint(AbstractEndpoint abstractEndpoint) {
        return this.subdomainName.equals(abstractEndpoint.getSubDomainName())
                || abstractEndpoint.getServiceName().getNamespaceURI().indexOf("private") <= 0;
    }

    /**
     * 
     * @param serviceEndpoint
     * @return
     */
    private boolean isNewEndpoint(ServiceEndpoint serviceEndpoint) {
        this.log.call();

        try {
            this.newEndpointContext.lookup(getKey(serviceEndpoint));
            return true;
        } catch (final NamingException e) {
            return false;
        }
    }

    /**
     * This method will update the given endpoint description and its
     * interfaces.
     * 
     * @param endpoint
     *            endpoint to update
     */
    private void updateEndpointDescriptionAndInterfaces(final AbstractEndpoint endpoint) {
        this.log.start();

        Document description = null;
        final List<QName> interfaces = new ArrayList<QName>(1);

        try {
            description = this.adminService.getServiceDescription(endpoint.getComponentName(),
                    endpoint);

            if (description == null) {
                this.log.warning("There is no WSDL description for the endpoint "
                        + endpoint.getServiceName() + "-" + endpoint.getEndpointName());
            } else {
                endpoint.setDescription(description);

                // get the interfaces
                final Description wsdlDescription = this.wsdlReader
                        .readWSDL4ComplexWSDL(description);
                if (wsdlDescription != null && wsdlDescription.getServices() != null) {
                    for (final Service service : wsdlDescription.getServices()) {
                        if (service.getEndpoint(endpoint.getEndpointName()) != null) {
                            if (service.getInterface() != null
                                    && service.getInterface().getQName() != null) {
                                // can't we have multiple interfaces for a
                                // service?
                                interfaces.add(service.getInterface().getQName());
                            }
                        }
                    }
                }

                if (interfaces.size() == 0) {
                    this.log.warning("No interface found in the WSDL description of the endpoint "
                            + endpoint.getServiceName() + "-" + endpoint.getEndpointName());
                }
            }
        } catch (final PetalsException e) {
            this.log.error("Failed to retrieve the WSDL description of the endpoint "
                    + endpoint.getServiceName() + "-" + endpoint.getEndpointName(), e);
        } catch (final WSDLException e) {
            this.log.error("Invalid description for the endpoint " + endpoint.getServiceName()
                    + "-" + endpoint.getEndpointName(), e);
        }

        if (interfaces.size() == 0) {
            final QName interfaze = new QName(endpoint.getServiceName().getNamespaceURI(),
                    UNRESOLVED_INTERFACE + endpoint.getServiceName().getLocalPart());
            this.log.warning("Set a default interface name for the endpoint "
                    + endpoint.getServiceName() + "-" + endpoint.getEndpointName() + " : "
                    + interfaze);
            interfaces.add(interfaze);
        }

        endpoint.setInterfaces(interfaces.toArray(new QName[0]));

        this.log.end();
    }

    /**
     * Get a unique key for the service endpoint
     * 
     * @param ep
     * @return
     */
    private static final String getKey(ServiceEndpoint ep) {
        return getKey(ep.getServiceName(), ep.getEndpointName());
    }

    /**
     * Create an unique key for the serviceName / endpointName
     * 
     * @return
     */
    private static final String getKey(QName serviceName, String endpointName) {
        final String result = serviceName + "@" + endpointName;
        return result.replaceAll("/", "&47;");
    }

    /**
     * Retrieve the service name from a jndi key
     * 
     * @return
     */
    private static final QName getServiceNameFromKey(String key) {
        final String[] result = key.split("@");
        return QName.valueOf(result[0].replaceAll("&47;", "/"));
    }

    /**
     * Retrieve the endpoint name from a jndi key
     * 
     * @return
     */
    private static final String getEndpointNameFromKey(String key) {
        final String[] result = key.split("@");
        return result[1].replaceAll("&47;", "/");
    }

    /**
     * Extract interface QName from the given endpoint.
     * 
     * @param abstractEndpoint
     * @return
     */
    private static final List<String> interfaceQNameToString(AbstractEndpoint abstractEndpoint) {
        final QName[] interfaces = abstractEndpoint.getInterfaces();
        final List<String> interfaceNames = new ArrayList<String>();
        if (interfaces != null && interfaces.length > 0) {
            for (final QName interfaceName : interfaces) {
                interfaceNames.add(interfaceName.toString());
            }
        } else {
            interfaceNames.add(UNDEFINED);
        }
        return interfaceNames;
    }

    public static void main(String[] args) throws Exception {
        final DocumentBuilder docb = DocumentBuilderFactory.newInstance().newDocumentBuilder();
        final Document description = docb.parse(new File("d:\\tmp\\test.wsdl"));
        final WSDLFactory wsdlFact = WSDLFactory.newInstance();
        final WSDLReader wsdlReader = wsdlFact.newWSDLReader();
        final Description desc = wsdlReader.readWSDL(description);
        final List<InterfaceType> interfaceTypes = desc.getInterfaces();
        for (final InterfaceType interfaceType : interfaceTypes) {
            System.out.println(interfaceType.getQName());
        }
    }
}
