/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.kernel.configuration;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.util.Iterator;
import java.util.Map;

import org.ow2.petals.jbi.messaging.routing.Router;
import org.ow2.petals.kernel.api.server.util.SystemUtil;
import org.ow2.petals.platform.repository.RepositoryService;
import org.ow2.petals.transport.Transporter;

/**
 * This class holds the overall configuration of a PEtALS container. It stores
 * both the local topology information and the global server properties
 * information.<br>
 * The configuration service retrieves the configuration when it is started and
 * creates an instance of this class. The instance is exposed on the distributed
 * JNDI server.
 * 
 * @author alouis, rnaudin - EBM WebSourcing
 * @author Mohamed Mokhtari
 * @since Petals 2.0
 * 
 */
public final class ContainerConfiguration implements Serializable {

    /**
     * the "unique" ID for serialization
     */
    private static final long serialVersionUID = 6543213857L;

    /**
     * Container state enumeration.
     */
    public static enum ContainerState {
        /**
         * Started state
         */
        STARTED,
        /**
         * Stopped state
         */
        STOPPED,
        /**
         * unknown state
         */
        UNKNOWN
    };

    // ********** Global ************//

    /**
     * Subdomain name.
     */
    private String subdomainName;

    /**
     * Container name.
     */
    private String name;

    /**
     * Container description
     */
    private String description;

    /**
     * Host of the container.
     */
    private String host;

    /**
     * State of the container.
     */
    private ContainerState state = ContainerState.UNKNOWN;

    /**
     * Install directory.
     */
    private String installDirectoryPath;

    /**
     * Repository directory path.
     */
    private String repositoryDirectoryPath;

    /**
     * Exchange validation.<br>
     * If <code>true</code>, used by the Address resolver to know if a component
     * is agree to receive a exchange from a consumer.
     */
    private boolean exchangeValidation;

    /**
     * Transport Quality of Service. {@see Transporter}
     */
    private String transportQOS = Transporter.DEFAULT_POLICY;

    /**
     * Routing Strategy. {@see EndpointChooserStrategy}
     */
    private String routingStrategy = Router.DEFAULT_PLATFORM_STRATEGY;

    /**
     * The timeout for processing the installation and deployment tasks.
     */
    private long taskTimeout;

    /**
     * Flag indicating if the JBI classLoaders must be isolated from the
     * container ClassLoader.
     */
    private boolean isolateJBIClassLoaders;

    /**
     * Flag indicating if the autoloader is activated or not.
     */
    private boolean activateAutoloader = true;

    /**
     * Login to connect to the node.
     */
    private String user;

    /**
     * Password to connect to the node.
     */
    private String password;

    // ********** Network ************//

    /**
     * Network port. Port used for PEtALS unicast messages.
     */
    private int networkBindingPort;

    // ************ JMX **************//

    /**
     * Port of the RMI adaptor for JMX.
     */
    private int jmxRMIConnectorPort;

    // ***** JORAM Transporter *****//

    /**
     * Port of the JORAM domain.
     */
    private int joramDomainPort;

    /**
     * Port of the JORAM TCP communication.
     */
    private int joramTCPPort;

    // ***** DREAM Transporter *****//

    /**
     * Port of the DREAM TCP communication.
     */
    private int dreamTCPPort;

    /**
     * Port of the Dream SSL communication.
     */
    private int dreamSSLPort;

    /**
     * The SSL key password
     */
    private String sslKeyPassword;

    /**
     * The SSL Keystore file
     */
    private String sslKeystore;

    /**
     * The SSL Keystore password
     */
    private String sslKeystorePassword;

    /**
     * The SSL Truststore file
     */
    private String sslTruststore;

    /**
     * The SSL Truststore password
     */
    private String sslTruststorePassword;

    /**
     * DREAM LFU cache size for TCP connections.
     */
    private int dreamSessionCacheSize = 20;

    /**
     * DREAM TCP connection timeout.
     */
    private int dreamConnectionTimeout = 15000;

    /**
     * DREAM TCP connection retry counter.
     */
    private int dreamConnectionRetry = 2;

    /**
     * The additional configuration in the server.properties defined by the
     * user. Get all the user.* key/value pairs.
     */
    private Map<String, String> userConfiguration;

    /**
     * 
     */
    public ContainerConfiguration() {
        super();
    }

    public void init() throws ConfigurationException {
        try {
            this.installDirectoryPath = SystemUtil.getPetalsInstallDirectory().getCanonicalPath();
        } catch (IOException e) {
            throw new ConfigurationException("The PEtALS install directory is not valid : "
                    + SystemUtil.getPetalsInstallDirectory().getAbsolutePath());
        }
        this.repositoryDirectoryPath = this.installDirectoryPath + File.separator
                + RepositoryService.DEFAULT_REPOSITORY_PATH;

    }

    public Map<String, String> getUserConfiguration() {
        return this.userConfiguration;
    }

    public String getSubdomainName() {
        return this.subdomainName;
    }

    public void setSubdomainName(final String subdomainName) {
        this.subdomainName = subdomainName;
    }

    public String getName() {
        return this.name;
    }

    public String getDescription() {
        return this.description;
    }

    public String getHost() {
        return this.host;
    }

    public ContainerState getState() {
        return this.state;
    }

    public String getInstallDirectoryPath() {
        return this.installDirectoryPath;
    }

    public String getRepositoryDirectoryPath() {
        return this.repositoryDirectoryPath;
    }

    public boolean isExchangeValidation() {
        return this.exchangeValidation;
    }

    public String getTransportQOS() {
        return this.transportQOS;
    }

    public String getRoutingStrategy() {
        return this.routingStrategy;
    }

    public long getTaskTimeout() {
        return this.taskTimeout;
    }

    public boolean isIsolateJBIClassLoaders() {
        return this.isolateJBIClassLoaders;
    }

    public boolean isActivateAutoloader() {
        return this.activateAutoloader;
    }

    public int getNetworkBindingPort() {
        return this.networkBindingPort;
    }

    public String getUser() {
        return this.user;
    }

    public String getPassword() {
        return this.password;
    }

    public int getJmxRMIConnectorPort() {
        return this.jmxRMIConnectorPort;
    }

    public int getJoramDomainPort() {
        return this.joramDomainPort;
    }

    public int getJoramTCPPort() {
        return this.joramTCPPort;
    }

    public int getDreamTCPPort() {
        return this.dreamTCPPort;
    }

    public int getDreamSSLPort() {
        return this.dreamSSLPort;
    }

    public String getSSLKeyPassword() {
        return this.sslKeyPassword;
    }

    public String getSSLKeystore() {
        return this.sslKeystore;
    }

    public String getSSLKeystorePassword() {
        return this.sslKeystorePassword;
    }

    public String getSSLTruststore() {
        return this.sslTruststore;
    }

    public String getSSLTruststorePassword() {
        return this.sslTruststorePassword;
    }

    public int getDreamSessionCacheSize() {
        return this.dreamSessionCacheSize;
    }

    public int getDreamConnectionTimeout() {
        return this.dreamConnectionTimeout;
    }

    public int getDreamConnectionRetry() {
        return this.dreamConnectionRetry;
    }

    public void setState(ContainerState state) {
        this.state = state;
    }

    @Override
    public String toString() {
        StringBuilder toStringBuilder = new StringBuilder();
        toStringBuilder.append("Container ").append(this.name).append(" on ").append(this.host)
                .append(" in sub-domain ").append(this.subdomainName);
        if (this.description != null) {
            toStringBuilder.append("\n\t\t\tDescription :\n\t\t\t  ").append(this.description);
        }
        toStringBuilder.append("\n\t\t\tConfiguration :\n\t\t\t  - InstallDirectory:").append(
                this.installDirectoryPath);
        toStringBuilder.append("\n\t\t\t  - RepositoryDirectory:").append(
                this.repositoryDirectoryPath);
        toStringBuilder.append("\n\t\t\t  - ExchangeValidation:").append(this.exchangeValidation);
        toStringBuilder.append("\n\t\t\t  - TransportQOS:").append(this.transportQOS);
        toStringBuilder.append("\n\t\t\t  - RoutingStrategy:").append(this.routingStrategy);
        toStringBuilder.append("\n\t\t\t  - TaskTimeout:").append(this.taskTimeout);
        toStringBuilder.append("\n\t\t\t  - IsolateClassLoaders:").append(
                this.isolateJBIClassLoaders);
        toStringBuilder.append("\n\t\t\t  - ActivateAutoloader:").append(this.activateAutoloader);
        toStringBuilder.append("\n\t\t\t  - User:")
                .append((this.user != null) ? this.user : "none");
        if (this.password != null) {
            StringBuffer hiddenPassword = new StringBuffer(this.password.length());
            while (hiddenPassword.length() < this.password.length()) {
                hiddenPassword.append("*");
            }
            toStringBuilder.append("\n\t\t\t  - Password:").append(hiddenPassword);
        } else {
            toStringBuilder.append("\n\t\t\t  - Password:none");
        }
        toStringBuilder.append("\n\t\t\t  - NetworkBindingPort:").append(
                (this.networkBindingPort != 0) ? this.networkBindingPort : "none");
        toStringBuilder.append("\n\t\t\t  - JMXRMIConnectionPort:")
                .append(this.jmxRMIConnectorPort);
        toStringBuilder.append("\n\t\t\t  - JoramDomainPort:").append(
                (this.joramDomainPort != 0) ? this.joramDomainPort : "none");
        toStringBuilder.append("\n\t\t\t  - JoramTCPPort:").append(
                (this.joramTCPPort != 0) ? this.joramTCPPort : "none");
        toStringBuilder.append("\n\t\t\t  - DreamTCPPort:").append(
                (this.dreamTCPPort != 0) ? this.dreamTCPPort : "none");
        toStringBuilder.append("\n\t\t\t  - DreamSSLPort:").append(
                (this.dreamSSLPort != 0) ? this.dreamSSLPort : "none");
        if (this.sslKeyPassword != null) {
            StringBuffer hiddenPassword = new StringBuffer(this.sslKeyPassword.length());
            while (hiddenPassword.length() < this.sslKeyPassword.length()) {
                hiddenPassword.append("*");
            }
            toStringBuilder.append("\n\t\t\t  - sslKeyPassword:").append(hiddenPassword);
        } else {
            toStringBuilder.append("\n\t\t\t  - sslKeyPassword:none");
        }
        toStringBuilder.append("\n\t\t\t  - Keystore:").append(
                (this.sslKeystore != null) ? this.sslKeystore : "none");
        if (this.sslKeystorePassword != null) {
            StringBuffer hiddenPassword = new StringBuffer(this.sslKeystorePassword.length());
            while (hiddenPassword.length() < this.sslKeystorePassword.length()) {
                hiddenPassword.append("*");
            }
            toStringBuilder.append("\n\t\t\t  - KeystorePassword:").append(hiddenPassword);
        } else {
            toStringBuilder.append("\n\t\t\t  - KeystorePassword:none");
        }
        toStringBuilder.append("\n\t\t\t  - Truststore:").append(
                (this.sslTruststore != null) ? this.sslTruststore : "none");
        if (this.sslTruststorePassword != null) {
            StringBuffer hiddenPassword = new StringBuffer(this.sslTruststorePassword.length());
            while (hiddenPassword.length() < this.sslTruststorePassword.length()) {
                hiddenPassword.append("*");
            }
            toStringBuilder.append("\n\t\t\t  - TruststorePassword:").append(hiddenPassword);
        } else {
            toStringBuilder.append("\n\t\t\t  - TruststorePassword:none");
        }
        toStringBuilder.append("\n\t\t\t  - DreamSessionCacheSize:").append(
                this.dreamSessionCacheSize);
        toStringBuilder.append("\n\t\t\t  - DreamConnectionTimeout:").append(
                this.dreamConnectionTimeout);
        toStringBuilder.append("\n\t\t\t  - DreamConnectionRetry:").append(
                this.dreamConnectionRetry);
        toStringBuilder.append("\n\t\t\t  - State:").append(this.state.name());

        if (this.userConfiguration != null) {
            StringBuffer userBuffer = new StringBuffer();
            Iterator<String> iter = this.userConfiguration.keySet().iterator();
            while (iter.hasNext()) {
                String key = iter.next();
                userBuffer.append("\n\t\t\t\t").append(key).append('=').append(
                        this.userConfiguration.get(key));
            }
            toStringBuilder.append("\n\t\t\t  - User properties:").append(userBuffer.toString());
        }

        return toStringBuilder.toString();
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + (activateAutoloader ? 1231 : 1237);
        result = prime * result + ((description == null) ? 0 : description.hashCode());
        result = prime * result + dreamConnectionRetry;
        result = prime * result + dreamConnectionTimeout;
        result = prime * result + dreamSSLPort;
        result = prime * result + dreamSessionCacheSize;
        result = prime * result + dreamTCPPort;
        result = prime * result + (exchangeValidation ? 1231 : 1237);
        result = prime * result + ((host == null) ? 0 : host.hashCode());
        result = prime * result
                + ((installDirectoryPath == null) ? 0 : installDirectoryPath.hashCode());
        result = prime * result + (isolateJBIClassLoaders ? 1231 : 1237);
        result = prime * result + jmxRMIConnectorPort;
        result = prime * result + joramDomainPort;
        result = prime * result + joramTCPPort;
        result = prime * result + ((sslKeyPassword == null) ? 0 : sslKeyPassword.hashCode());
        result = prime * result + ((sslKeystore == null) ? 0 : sslKeystore.hashCode());
        result = prime * result
                + ((sslKeystorePassword == null) ? 0 : sslKeystorePassword.hashCode());
        result = prime * result + ((name == null) ? 0 : name.hashCode());
        result = prime * result + networkBindingPort;
        result = prime * result + ((password == null) ? 0 : password.hashCode());
        result = prime * result
                + ((repositoryDirectoryPath == null) ? 0 : repositoryDirectoryPath.hashCode());
        result = prime * result + ((routingStrategy == null) ? 0 : routingStrategy.hashCode());
        result = prime * result + ((state == null) ? 0 : state.hashCode());
        result = prime * result + ((subdomainName == null) ? 0 : subdomainName.hashCode());
        result = prime * result + (int) (taskTimeout ^ (taskTimeout >>> 32));
        result = prime * result + ((transportQOS == null) ? 0 : transportQOS.hashCode());
        result = prime * result + ((sslTruststore == null) ? 0 : sslTruststore.hashCode());
        result = prime * result
                + ((sslTruststorePassword == null) ? 0 : sslTruststorePassword.hashCode());
        result = prime * result + ((user == null) ? 0 : user.hashCode());
        result = prime * result + ((userConfiguration == null) ? 0 : userConfiguration.hashCode());
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        boolean result = false;
        if (obj instanceof ContainerConfiguration) {
            ContainerConfiguration otherContainerConfiguration = (ContainerConfiguration) obj;
            result = (otherContainerConfiguration.name.equals(this.name)
                    && otherContainerConfiguration.host.equals(this.host)
                    && otherContainerConfiguration.exchangeValidation == this.exchangeValidation
                    && otherContainerConfiguration.transportQOS.equals(this.transportQOS)
                    && otherContainerConfiguration.routingStrategy.equals(this.routingStrategy)
                    && otherContainerConfiguration.taskTimeout == this.taskTimeout
                    && otherContainerConfiguration.isolateJBIClassLoaders == this.isolateJBIClassLoaders
                    && otherContainerConfiguration.activateAutoloader == this.activateAutoloader
                    && otherContainerConfiguration.networkBindingPort == this.networkBindingPort
                    && otherContainerConfiguration.user.equals(this.user)
                    && otherContainerConfiguration.password.equals(this.password)
                    && otherContainerConfiguration.jmxRMIConnectorPort == this.jmxRMIConnectorPort
                    && otherContainerConfiguration.joramDomainPort == this.joramDomainPort
                    && otherContainerConfiguration.joramTCPPort == this.joramTCPPort
                    && otherContainerConfiguration.dreamTCPPort == this.dreamTCPPort
                    && otherContainerConfiguration.dreamSSLPort == this.dreamSSLPort
                    && otherContainerConfiguration.sslKeyPassword.equals(this.sslKeyPassword)
                    && otherContainerConfiguration.sslKeystore.equals(this.sslKeystore)
                    && otherContainerConfiguration.sslKeystorePassword
                            .equals(this.sslKeystorePassword)
                    && otherContainerConfiguration.sslTruststore.equals(this.sslTruststore)
                    && otherContainerConfiguration.sslTruststorePassword
                            .equals(this.sslTruststorePassword) && (otherContainerConfiguration.userConfiguration == this.userConfiguration || (otherContainerConfiguration.userConfiguration != null && otherContainerConfiguration.userConfiguration
                    .equals(this.userConfiguration))));
        }
        return result;
    }

    protected void setUserConfiguration(Map<String, String> userConfiguration) {
        this.userConfiguration = userConfiguration;
    }

    protected void setName(String name) {
        this.name = name;
    }

    protected void setHost(String host) {
        this.host = host;
    }

    protected void setDescription(String description) {
        this.description = description;
    }

    protected void setInstallDirectoryPath(String installDirectoryPath) {
        this.installDirectoryPath = installDirectoryPath;
    }

    protected void setRepositoryDirectoryPath(String repositoryDirectoryPath) {
        this.repositoryDirectoryPath = repositoryDirectoryPath;
    }

    protected void setExchangeValidation(boolean exchangeValidation) {
        this.exchangeValidation = exchangeValidation;
    }

    protected void setTransportQOS(String transportQOS) {
        this.transportQOS = transportQOS;
    }

    protected void setRoutingStrategy(String routingStrategy) {
        this.routingStrategy = routingStrategy;
    }

    protected void setTaskTimeout(long taskTimeout) {
        this.taskTimeout = taskTimeout;
    }

    protected void setIsolateJBIClassLoaders(boolean isolatedJBIClassLoaders) {
        this.isolateJBIClassLoaders = isolatedJBIClassLoaders;
    }

    protected void setActivateAutoloader(boolean activateAutoloader) {
        this.activateAutoloader = activateAutoloader;
    }

    protected void setNetworkBindingPort(int networkBindingPort) {
        this.networkBindingPort = networkBindingPort;
    }

    protected void setUser(String user) {
        this.user = user;
    }

    protected void setPassword(String password) {
        this.password = password;
    }

    protected void setJmxRMIConnectorPort(int jmxRMIConnectorPort) {
        this.jmxRMIConnectorPort = jmxRMIConnectorPort;
    }

    protected void setJoramDomainPort(int joramDomainPort) {
        this.joramDomainPort = joramDomainPort;
    }

    protected void setJoramTCPPort(int joramTCPPort) {
        this.joramTCPPort = joramTCPPort;
    }

    protected void setDreamTCPPort(int dreamTCPPort) {
        this.dreamTCPPort = dreamTCPPort;
    }

    protected void setDreamSSLPort(int dreamSSLPort) {
        this.dreamSSLPort = dreamSSLPort;
    }

    protected void setSSLKeyPassword(String keyPassword) {
        this.sslKeyPassword = keyPassword;
    }

    protected void setSSLKeystore(String keystore) {
        this.sslKeystore = keystore;
    }

    protected void setSSLKeystorePassword(String keystorePassword) {
        this.sslKeystorePassword = keystorePassword;
    }

    protected void setSSLTruststore(String truststore) {
        this.sslTruststore = truststore;
    }

    protected void setSSLTruststorePassword(String truststorePassword) {
        this.sslTruststorePassword = truststorePassword;
    }

    protected void setDreamSessionCacheSize(int dreamSessionCacheSize) {
        this.dreamSessionCacheSize = dreamSessionCacheSize;
    }

    protected void setDreamConnectionTimeout(int dreamConnectionTimeout) {
        this.dreamConnectionTimeout = dreamConnectionTimeout;
    }

    protected void setDreamConnectionRetry(int dreamConnectionRetry) {
        this.dreamConnectionRetry = dreamConnectionRetry;
    }

}
