/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005-2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.kernel.server;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import java.util.Properties;
import java.util.Vector;

import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalContentException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.util.Fractal;
import org.objectweb.util.monolog.Monolog;
import org.ow2.petals.communication.jndi.client.JNDIService;
import org.ow2.petals.communication.topology.TopologyService;
import org.ow2.petals.jbi.management.ManagementException;
import org.ow2.petals.jbi.management.recovery.SystemRecoveryService;
import org.ow2.petals.kernel.admin.PetalsAdminServiceInterface;
import org.ow2.petals.kernel.api.server.PetalsException;
import org.ow2.petals.kernel.api.server.PetalsServer;
import org.ow2.petals.kernel.api.server.PetalsStateListener;
import org.ow2.petals.kernel.api.server.util.SystemUtil;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.kernel.configuration.ContainerConfiguration;
import org.ow2.petals.kernel.configuration.DomainConfiguration;
import org.ow2.petals.transport.Transporter;
import org.ow2.petals.transport.util.DreamUtil;
import org.ow2.petals.util.JNDIUtil;

import static org.ow2.petals.kernel.server.FractalHelper.AUTOLOADER_COMPONENT;
import static org.ow2.petals.kernel.server.FractalHelper.COMM_COMPOSITE;
import static org.ow2.petals.kernel.server.FractalHelper.CONFIGURATION_COMPONENT;

/**
 * This class allows to start and stop the Petals Server.
 * 
 * @author rnaudin - eBMWebsourcing
 * 
 */
public class PetalsServerImpl implements PetalsServer {

    private static final String LOGGER_PROPERTIES_FILE_NAME = "loggers.properties";

    private static final String[] MONOLOG_FILE_HANDLERS = { "petalsFile", "joramFile", "dreamFile" };

    /**
     * Listener called back when petals is started or stopped
     */
    private List<PetalsStateListener> petalsListeners;

    /**
     * Thread used to stop Petals in an isolate thread
     */
    private PetalsStopThread petalsStopThread;

    /**
     * The domain configuration
     */
    private DomainConfiguration domainConfiguration;

    /**
     * The container configuration
     */
    private ContainerConfiguration containerConfiguration;

    /**
     * Petals composite
     */
    private Component petalsComposite;

    /**
     * Petals Content Controller
     */
    private ContentController petalsContentController;

    /**
     * Creates a new instance of {@link PetalsServerImpl}
     * 
     * @throws PetalsException
     */
    public PetalsServerImpl() throws PetalsException {
        this(false);
    }

    /**
     * Creates a new instance of {@link PetalsServerImpl}
     * 
     * @throws PetalsException
     */
    public PetalsServerImpl(boolean observer) throws PetalsException {
        this.petalsStopThread = new PetalsStopThread(this);
        this.petalsListeners = new Vector<PetalsStateListener>();
    }

    /**
     * Initialize the PEtALS environment. Init Qname, JMX, Monolog, Fractal.
     * 
     * @see org.ow2.petals.kernel.api.server.PetalsServer#init()
     */
    public void init() throws PetalsException {

        // set the compatible SUID for the Qname
        System.setProperty("com.sun.xml.namespace.QName.useCompatibleSerialVersionUID", "1.0");

        // set the JMX server
        System
                .setProperty("javax.management.builder.initial",
                        "mx4j.server.MX4JMBeanServerBuilder");

        try {
            // initialize MONOLOG
            this.initializeMonolog();

            // initialize Petals Fractal composite
            this.initializePetalsComposite();

        } catch (IOException exception) {
            // clean up the petals server
            throw new PetalsException("Problem while initializing Petals", exception);
        }
    }

    /**
     * Start the Petals Server. Start the fractal components. Register the
     * different component, then try to recover the Petals container.
     * 
     * @see org.ow2.petals.kernel.api.server.PetalsServer#start()
     */
    public void start() throws PetalsException {

        try {
            // start PEtALS Fractal composite
            this.startPetalsComposite();

            // register the PEtALS server in the Petals admin.
            this.registerPetalsServer();

            // expose PEtALS MBean object into the JMX server
            MBeanHelper.registerMBeans(this.petalsComposite);

            // recover all JBI entities
            this.recoverSystem();

            // start the transporter traffics once the recovery of SAs is done
            this.startTransporterTraffics();

            this.startupDone();

        } catch (Exception exception) {
            // clean up the petals server
            if (this.petalsComposite != null) {
                System.err.println("Problem while starting PEtALS, try to stop PEtALS cleanly...");
                try {
                    FractalHelper.stopComponent(this.petalsComposite);
                } catch (Throwable e) {
                    System.err.println("Failed to stop PEtALS cleanly");
                }
            }
            throw new PetalsException("Problem while starting PEtALS", exception);
        }

        // notify the listener if there is one
        for (PetalsStateListener petalsListener : this.petalsListeners) {
            petalsListener.onPetalsStarted();
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.kernel.api.server.PetalsServer#stop()
     */
    public void stop() throws PetalsException {
        Exception exception = null;
        try {
            this.stopPetalsComposite();
        } catch (Exception e) {
            exception = e;
        }

        // notify the listener if there is one
        for (PetalsStateListener petalsListener : this.petalsListeners) {
            if (exception == null) {
                petalsListener.onPetalsStopped(true);
            } else {
                petalsListener.onPetalsStopped(false);
            }
        }

        // NOTE : unreachable part if the listener exists the system

        if (exception != null) {
            if (exception instanceof PetalsException) {
                throw (PetalsException) exception;
            } else {
                throw new PetalsException(exception);
            }

        }

    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.kernel.api.server.PetalsServer#addPetalsStateListener(
     * org.ow2.petals.kernel.api.server.PetalsStateListener)
     */
    public void addPetalsStateListener(PetalsStateListener petalsListener) {
        this.petalsListeners.add(petalsListener);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.kernel.api.server.PetalsServer#removePetalsStateListener
     * (org.ow2.petals.kernel.api.server.PetalsStateListener)
     */
    public boolean removePetalsStateListener(PetalsStateListener petalsListener) {
        return this.petalsListeners.remove(petalsListener);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.kernel.api.server.PetalsServer#getContainerConfiguration()
     */
    public String getContainerConfiguration() throws PetalsException {
        if (this.containerConfiguration == null) {
            throw new PetalsException("The container configuration is not properly set");
        }

        return this.containerConfiguration.toString();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.kernel.api.server.PetalsServer#browseJNDI()
     */
    public String browseJNDI() throws PetalsException {
        if (this.containerConfiguration == null) {
            throw new PetalsException("The container configuration is not properly set");
        }

        String result = null;
        Component jndiComponent = FractalHelper.getRecursiveComponentByName(
                this.petalsContentController, FractalHelper.JNDI_COMPONENT);
        try {
            JNDIService jndiService = (JNDIService) jndiComponent.getFcInterface("service");
            InitialContext initialContext = jndiService.getInitialContext();
            if (this.domainConfiguration.getJndiHost() == null) {
                result = JNDIUtil.browseJNDI(initialContext, (String) null, 0);
            } else {
                result = JNDIUtil.browseJNDI(initialContext,
                        this.domainConfiguration.getJndiHost(), this.domainConfiguration
                                .getJndiPort());
            }

        } catch (NoSuchInterfaceException e) {
            throw new PetalsException(e);
        } catch (NamingException e) {
            throw new PetalsException(e);
        }

        return result;
    }

    /**
     * Initialize petals composite
     * 
     * @throws PetalsException
     * 
     */
    private void initializePetalsComposite() throws PetalsException {
        try {
            this.petalsComposite = FractalHelper.createNewComponent(FractalHelper.PETALS_COMPOSITE);
            this.petalsContentController = Fractal.getContentController(this.petalsComposite);
        } catch (ADLException e) {
            throw new PetalsException("Error creating PEtALS main Composite", e);
        } catch (NoSuchInterfaceException e) {
            throw new PetalsException("Error creating PEtALS main Composite", e);
        }
    }

    /**
     * Initialize the environment for Monolog and start the Monolog factory
     * 
     * @throws IOException
     */
    private void initializeMonolog() throws IOException {
        // get the loggers properties
        URL logConfUrl = this.getClass().getResource("/" + LOGGER_PROPERTIES_FILE_NAME);
        if (logConfUrl == null) {
            throw new IOException("Failed to reach Monolog resource '"
                    + LOGGER_PROPERTIES_FILE_NAME + "'");
        }
        Properties logProperties = new Properties();
        logProperties.load(logConfUrl.openStream());

        // configure the Monolog handlers
        this.configureMonologHandlers(logProperties);

        Monolog.getMonologFactory(logProperties);
    }

    /**
     * Configure the Monolog handlers
     * 
     * @param logProperties
     */
    private void configureMonologHandlers(Properties logProperties) {
        // create logs directory is not present
        File logDir = new File(SystemUtil.getPetalsInstallDirectory(), "logs");
        logDir.mkdirs();

        // get the current date as a string
        Date currentDate = new Date(System.currentTimeMillis());
        SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd");
        String dateString = dateFormat.format(currentDate);

        for (String handlerName : MONOLOG_FILE_HANDLERS) {
            // set the path names of the Monologs file handlers
            String handlerOutput = logProperties.getProperty("handler." + handlerName + ".output");
            if (handlerOutput != null) {
                handlerOutput = handlerOutput.replace("%d", dateString);
                logProperties.setProperty("handler." + handlerName + ".output", new File(logDir,
                        handlerOutput).getAbsolutePath());
            }

            // TODO: Monolog bug on the appendMode field
            String handlerAppendMode = logProperties.getProperty("handler." + handlerName
                    + ".appendMode");
            if ("true".equals(handlerAppendMode) || "false".equals(handlerAppendMode)) {
                System.setProperty(handlerAppendMode, handlerAppendMode);
            }
        }
    }

    /**
     * Start the Petals Fractal composite. TODO create a Fractal component to
     * handle the starting and stoping of Fractal composites/components.
     * 
     * @throws NoSuchInterfaceException
     * @throws IllegalLifeCycleException
     * @throws ADLException
     * @throws IllegalBindingException
     * @throws IllegalContentException
     * @throws PetalsException
     */
    private void startPetalsComposite() throws NoSuchInterfaceException, IllegalLifeCycleException,
            ADLException, IllegalContentException, IllegalBindingException, PetalsException {

        // start first the configuration component
        Component configurationComponent = FractalHelper.getComponentByName(
                this.petalsContentController, CONFIGURATION_COMPONENT);
        if (!FractalHelper.startComponent(configurationComponent)) {
            throw new PetalsException("Failed to start PEtALS Fractal component "
                    + CONFIGURATION_COMPONENT);
        }

        // retrieve the configuration from the Configuration Service
        ConfigurationService configurationService = (ConfigurationService) configurationComponent
                .getFcInterface("service");
        this.containerConfiguration = configurationService.getContainerConfiguration();
        this.domainConfiguration = configurationService.getDomainConfiguration();
        
        // remove the autoloader component if not activated
        if (!this.containerConfiguration.isActivateAutoloader()) {
            Component autoloaderComponent = FractalHelper.getComponentByName(
                    this.petalsContentController, AUTOLOADER_COMPONENT);
            BindingController autoLoaderBC = Fractal.getBindingController(autoloaderComponent);
            for (String bindingName : autoLoaderBC.listFc()) {
                autoLoaderBC.unbindFc(bindingName);
            }
            this.petalsContentController.removeFcSubComponent(autoloaderComponent);
        }

        // then update the TCPPort fractal attribute of the DreamTransporter
        if (!DomainConfiguration.DomainMode.STANDALONE.equals(this.domainConfiguration.getMode())) {
            DreamUtil.setupDreamTransporter(this.petalsContentController,
                    this.containerConfiguration);
        }
        
        // start the communication composite
        Component communicationComposite = FractalHelper.getRecursiveComponentByName(
                this.petalsContentController, COMM_COMPOSITE);
        if (!FractalHelper.startComponent(communicationComposite)) {
            throw new PetalsException("Failed to start PEtALS Fractal composite " + COMM_COMPOSITE);
        }

        // finally, start the others components
        if (!FractalHelper.startComponent(this.petalsComposite)) {
            throw new PetalsException("Failed to start PEtALS Fractal components");
        }
    }

    /**
     * Stop the Petals Fractal composite.
     * 
     * @throws Exception
     */
    private void stopPetalsComposite() throws Exception {

        LifeCycleController lifeCycleController = Fractal
                .getLifeCycleController(this.petalsComposite);

        if (LifeCycleController.STARTED.equals(lifeCycleController.getFcState())) {

            if (DomainConfiguration.DomainMode.STANDALONE
                    .equals(this.domainConfiguration.getMode())) {
                // Standalone mode, prepare the stop of ongoing transfers
                Component standaloneTransporterComponent = FractalHelper
                        .getRecursiveComponentByName(this.petalsContentController,
                                FractalHelper.STANDALONE_TRANSPORTER_COMPONENT);
                Transporter standaloneTransporter = (Transporter) standaloneTransporterComponent
                        .getFcInterface("service");
                standaloneTransporter.stopTraffic();
            } else {
                // Platform (distributed) mode
                // First prepare the stop of the Dispatcher, to stop the ongoing
                // transfers
                Component dispatcherComponent = FractalHelper.getRecursiveComponentByName(
                        this.petalsContentController, FractalHelper.DISPATCHER_COMPONENT);
                Transporter dispatcher = (Transporter) dispatcherComponent
                        .getFcInterface("service");
                dispatcher.stopTraffic();

                // Then stop the Transport Protocols, to avoid blocking calls to
                // the interface TransportProtocolListener.
                Component joramComponent = FractalHelper.getRecursiveComponentByName(
                        this.petalsContentController, FractalHelper.JORAMTRANSPORTER_COMPONENT);
                LifeCycleController joramLCController = Fractal
                        .getLifeCycleController(joramComponent);
                joramLCController.stopFc();

                Component localComponent = FractalHelper.getRecursiveComponentByName(
                        this.petalsContentController, FractalHelper.MEMORYTRANSPORTER_COMPONENT);
                LifeCycleController localLCController = Fractal
                        .getLifeCycleController(localComponent);
                localLCController.stopFc();

                Component dreamComponent = FractalHelper.getRecursiveComponentByName(
                        this.petalsContentController, FractalHelper.DREAMTRANSPORTER_COMPONENT);
                LifeCycleController dreamLCController = Fractal
                        .getLifeCycleController(dreamComponent);
                dreamLCController.stopFc();
            }

            // Finally, stop the others components, we must stop the Composite
            // elements in an opposite way than their start order
            FractalHelper.stopComposite(this.petalsComposite);
        }
    }

    /**
     * Register to the PetalsAdmin service this Petals server instance
     * 
     * @throws NoSuchInterfaceException
     * @throws PetalsException
     */
    private void registerPetalsServer() throws NoSuchInterfaceException, PetalsException {
        Component petalsAdminComponent = FractalHelper.getRecursiveComponentByName(
                this.petalsContentController, FractalHelper.PETALSADMIN_COMPONENT);
        PetalsAdminServiceInterface petalsAdminService = (PetalsAdminServiceInterface) petalsAdminComponent
                .getFcInterface("service");

        petalsAdminService.setPetalsStopThread(this.petalsStopThread);
    }

    /**
     * Recover the system state. All state elements are recovered (components,
     * share libraries, service assemblies).
     * 
     * @throws NoSuchInterfaceException
     * @throws ManagementException
     * @throws ADLException
     * @throws PetalsException
     */
    private void recoverSystem() throws NoSuchInterfaceException, ManagementException,
            ADLException, PetalsException {
        Component systemRecoveryComponent = FractalHelper.getRecursiveComponentByName(
                this.petalsContentController, FractalHelper.SYSTEMRECOVERY_COMPONENT);
        SystemRecoveryService systemRecoveryService = (SystemRecoveryService) systemRecoveryComponent
                .getFcInterface("service");

        systemRecoveryService.recoverAllEntities();
    }

    /**
     * Start the transporter traffics.
     * 
     * @throws NoSuchInterfaceException
     */
    private void startTransporterTraffics() throws NoSuchInterfaceException {
        if (!DomainConfiguration.DomainMode.STANDALONE.equals(this.domainConfiguration.getMode())) {
            // Start only for the Platform (distributed) mode
            Component dispatcherComponent = FractalHelper.getRecursiveComponentByName(
                    this.petalsContentController, FractalHelper.DISPATCHER_COMPONENT);
            Transporter dispatcher = (Transporter) dispatcherComponent.getFcInterface("service");
            dispatcher.startTraffic();
        }
    }

    /**
     * Set the local container state to started. The local container state is in
     * unknown state until all is really started.
     * 
     * @throws Exception
     */
    private void startupDone() throws Exception {
        Component topologyComponent = FractalHelper.getRecursiveComponentByName(
                this.petalsContentController, FractalHelper.TOPOLOGY_COMPONENT);

        if (topologyComponent == null) {
            if (!DomainConfiguration.DomainMode.STANDALONE.equals(this.domainConfiguration
                    .getMode())) {
                throw new NullPointerException("The network fractal component is null");
            }
        } else {
            TopologyService service = (TopologyService) topologyComponent.getFcInterface("service");
            service.setContainerState(this.containerConfiguration.getName(),
                    ContainerConfiguration.ContainerState.STARTED);
        }
    }
}
