/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 eBM Websourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: RepositoryService.java 7970 2008-04-25 13:49:19Z rnaudin $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.platform.repository;

import java.io.File;
import java.net.URI;

import org.ow2.petals.kernel.api.server.PetalsException;

/**
 * Platform Component Repository.
 *
 * @version $Rev: 7970 $ $Date: 2008-04-25 15:49:19 +0200 (Fri, 25 Apr 2008) $
 * @since Petals 1.0
 * @author ofabre - eBM Websourcing
 */
public interface RepositoryService {

    /**
     * Path to repository location within Petals installation directory.
     */
    static final String DEFAULT_REPOSITORY_PATH = "repository";

    /**
     * Copy a component package into petals repository
     *
     * @param componentID
     *            <code>String</code> unique component id that identifies an
     *            unique component directory into the petals repository. Must be
     *            non null and non empty
     * @param tempFile
     *            <code>File</code> the component package file to add into the
     *            repository. Must be non null
     * @return the component installation root {@link URI}
     * @throws PetalsException
     *             if it fails to create install and work dirs for the component
     *             or if it fails to copy package into install dir
     */
    URI addComponentPackage(String componentID, File tempFile)
            throws PetalsException;

    /**
     * Copy a shared library package into petals repository
     *
     * @param sharedLibID
     *            <code>String</code> unique sl id that identifies an unique
     *            sl directory into the petals repository. Must be non null and
     *            non empty
     * @param tempFile
     *            <code>File</code> the shared lib package file to add into
     *            the repository. Must be non null
     * @return the sl installation root {@link URI}
     * @throws PetalsException
     *             if it fails to create install and work dirs for the sl or if
     *             it fails to copy package into install dir
     */
    URI addSharedLibPackage(String sharedLibID, File tempFile)
            throws PetalsException;

    /**
     * Copy a service assembly package into petals repository
     *
     * @param serviceAssemblyID
     *            <code>String</code> unique sa id that identifies an unique
     *            sa directory into the petals repository. Must be non null and
     *            non empty
     * @param tempFile
     *            <code>File</code> the service assembly package file to add
     *            into the repository. Must be non null
     * @return the sa installation root {@link URI}
     * @throws PetalsException
     *             if it fails to create install and work dirs for the sa or if
     *             it fails to copy package into install dir
     */
    URI addServiceAssemblyPackage(String serviceAssemblyID, File tempFile)
            throws PetalsException;

    /**
     * Explode a given service unit archive into a SU install root created into
     * the SA install root
     *
     * @param suId
     *            <code>String</code> unique SU identifier within the
     *            repository.
     * @param suZipLocation
     *            <code>URI</code> the SU Zip archive file to add flatten into
     *            the repository.
     * @param saId
     *            <code>String</code> unique SA identifier within the
     *            repository where SU add to be flatten
     * @return <code>URI</code> installation root of su in the repository.
     *         Must be a valid directory.
     * @throws PetalsException
     *             if SU install root can't be created or SU archive can't be
     *             unzipped
     */
    URI explodeSUIntoSAInstallFolder(String suId, URI suZipLocation, String saId)
            throws PetalsException;

    /**
     * Returns the install directory of the component
     *
     * @param componentId
     *            <code>String</code> unique component id that identifies an
     *            unique component directory into the petals repository.
     * @return the install directory of the component
     * @throws PetalsException
     *             if an error occured during component root retrieval
     */
    File getComponentInstallRoot(String componentId) throws PetalsException;

    /**
     * Returns the root directory of this component
     *
     * @param componentId
     *            <code>String</code> unique component id that identifies an
     *            unique component directory into the petals repository.
     * @return the root directory of this component, null if it doesn't exist
     * @throws PetalsException
     *             if an error occured during component root retrieval
     */
    File getComponentRoot(String componentId) throws PetalsException;

    /**
     * Returns the work directory of the component
     *
     * @param componentId
     *            <code>String</code> unique component id that identifies an
     *            unique component directory into the petals repository.
     * @return the work directory of the component
     * @throws PetalsException
     *             if an error occured during component root retrieval
     */
    File getComponentWorkRoot(String componentId) throws PetalsException;

    /**
     * Returns the root directory of this sl
     *
     * @param slId
     *            <code>String</code> unique sl id that identifies an unique
     *            sl directory into the petals repository.
     * @return the root directory of this sl, null if it doesn't exist
     * @throws PetalsException
     *             if an error occured during shared lib root retrieval
     */
    File getSharedLibRoot(String slId) throws PetalsException;

    /**
     * Returns the work directory of the sl
     *
     * @param slId
     *            <code>String</code> unique sl id that identifies an unique
     *            sl directory into the petals repository.
     * @return the work directory of the sl
     * @throws PetalsException
     *             if an error occured during shared lib root retrieval
     */
    File getSharedLibWorkRoot(String slId) throws PetalsException;

    /**
     * Returns the install directory of the sl
     *
     * @param slId
     *            <code>String</code> unique sl id that identifies an unique
     *            sl directory into the petals repository.
     * @return the install directory of the sl
     * @throws PetalsException
     *             if an error occured during shared lib root retrieval
     */
    File getSharedLibInstallRoot(String slId) throws PetalsException;

    /**
     * Returns the root directory of this sa
     *
     * @param saId
     *            <code>String</code> unique sa id that identifies an unique
     *            sa directory into the petals repository.
     * @return the root directory of this sa, null if it doesn't exist
     * @throws PetalsException
     *             if an error occured during service assembly root retrieval
     */
    File getServiceAssemblyRoot(String saId) throws PetalsException;

    /**
     * Returns the work directory of the sa
     *
     * @param saId
     *            <code>String</code> unique sa id that identifies an unique
     *            sa directory into the petals repository.
     * @return the work directory of the sa
     * @throws PetalsException
     *             if an error occured during service assembly root retrieval
     */
    File getServiceAssemblyWorkRoot(String saId) throws PetalsException;

    /**
     * Returns the install directory of the sa
     *
     * @param saId
     *            <code>String</code> unique sa id that identifies an unique
     *            sa directory into the petals repository.
     * @return the install directory of the sa
     * @throws PetalsException
     *             if an error occured during service assembly root retrieval
     */
    File getServiceAssemblyInstallRoot(String saId) throws PetalsException;

    /**
     * Delete and clean all directories and related data from the Petals
     * repository, for the given component identifier.
     *
     * @param componentId
     *            <code>String</code> unique component id that identifies an
     *            unique component directory into the petals repository.
     * @return <code>true</code> for sucessful removal. Otherwise,
     *         <code>false</code>.
     * @throws PetalsException
     *             if an error occured during component root retrieval
     */
    boolean removeComponentPackage(String componentId) throws PetalsException;

    /**
     * Delete and clean all directories and related data from the Petals
     * repository, for the given sl identifier.
     *
     * @param slId
     *            <code>String</code> unique sl id that identifies an unique
     *            sl directory into the petals repository.
     * @return <code>true</code> for sucessful removal. Otherwise,
     *         <code>false</code>.
     * @throws PetalsException
     *             if an error occured during shared lib root retrieval
     */
    boolean removeSharedLibPackage(String slId) throws PetalsException;

    /**
     * Delete and clean all directories and related data from the Petals
     * repository, for the given sa identifier.
     *
     * @param saId
     *            <code>String</code> unique sa id that identifies an unique
     *            sa directory into the petals repository.
     * @return <code>true</code> for sucessful removal. Otherwise,
     *         <code>false</code>.
     * @throws PetalsException
     *             if an error occured during service assembly root retrieval
     */
    boolean removeServiceAssemblyPackage(String saId) throws PetalsException;

    /**
     * Retrieve the components directory absolute path from the Petals
     * repository. This directory contains all installed components.
     *
     * @return the components directory URL from the Petals repository
     */
    File getComponentsDirectory();

    /**
     * Retrieve the ServiceAssemblies directory absolute path from the Petals
     * repository. This directory contains all deployed ServiceAssemblies.
     *
     * @return the ServiceAssemblies directory URL from the Petals repository

     */
    File getServiceAssembliesDirectory();

    /**
     * Retrieve the SharedLibs directory absolute path from the Petals
     * repository. This directory contains all installed SharedLibs.
     *
     * @return the SharedLibs directory URL from the Petals repository

     */
    File getSharedLibsDirectory();

    /**
     * Retrieve the "lost+found" directory absolute path from the Petals
     * repository. This directory contains all unknown SharedLibs, Components
     * and ServiceAssemblies.
     *
     * @return the SharedLibs directory URL from the Petals repository
     * @throws RuntimeException
     *             if an error occured during Petals root retrieval
     */
    String getLostPlusFoundDirectory();

    /**
     * Return petals repository folder. The repository folder contains extracted
     * component, service assembly and shared library packages. It will be
     * created when this method is called for the first time
     *
     * @return petals repository folder
     * @throws RuntimeException
     *             if the specified or default repository can't be created
     */
    File getRepositoryDirectory();
}
