/**
 * PETALS: PETALS Services Platform Copyright (C) 2005 EBM WebSourcing
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 * 
 * Initial developer(s): pierre.garcia@inrialpes.fr
 */

package org.ow2.petals.transport.util;

import java.util.HashMap;
import java.util.Map;

import javax.jbi.JBIException;

import org.objectweb.dream.message.ChunkAlreadyExistException;
import org.objectweb.dream.message.ChunkFactoryReference;
import org.objectweb.dream.message.InvalidReference;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManagerType;
import org.objectweb.dream.protocol.ExportIdentifierChunk;
import org.objectweb.dream.protocol.IPExportIdentifier;
import org.objectweb.dream.protocol.ProtocolChunk;
import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalContentException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.NameController;
import org.objectweb.fractal.util.Fractal;
import org.ow2.petals.communication.topology.TopologyService;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl;
import org.ow2.petals.kernel.api.server.PetalsException;
import org.ow2.petals.kernel.configuration.ContainerConfiguration;
import org.ow2.petals.kernel.server.FractalHelper;
import org.ow2.petals.transport.TransportException;
import org.ow2.petals.transport.Transporter;
import org.ow2.petals.transport.platform.dream.MessageExchangeChunk;
import org.ow2.petals.util.LoggingUtil;

/**
 * Dream utility.
 * 
 * @author pgarcia,MOKHTARI Mohamed
 */
public class DreamUtil {

    /**
     * Name of the chunk including destination
     */
    protected static final String DESTINATION_CHUNK_NAME = "destination";

    /**
     * Message Manager. Used here only to manage chunk in a message
     */
    protected MessageManagerType messageManagerItf;

    /**
     * Factory used to create Export Identifier Chunk
     */
    protected ChunkFactoryReference<ExportIdentifierChunk> exportIdChunkFactory;

    /**
     * Factory used to create, delete, copy... MessageExchangeChunk
     */
    ChunkFactoryReference<MessageExchangeChunk> messageExchangeChunkFactory;

    /**
     * Factory used to create, delete, copy... ProtocolChunk
     */
    ChunkFactoryReference<ProtocolChunk> protocolChunkFactory;

    /**
     * Inherited Logger
     */
    LoggingUtil log;

    /**
     * The Topology server
     */
    protected TopologyService topologyService;

    /**
     * Mapping between container name and ipExportChunk
     */
    protected static Map<String, Map<String, ExportIdentifierChunk>> exportIdChunks;

    protected static String defaultQOS;

    public DreamUtil(MessageManagerType messageManager, LoggingUtil log,
            TopologyService topologyService) {
        this.log = log;
        this.log.start();

        this.messageManagerItf = messageManager;
        // create the chunkFactory
        this.exportIdChunkFactory = this.messageManagerItf
                .getChunkFactory(ExportIdentifierChunk.class);
        this.messageExchangeChunkFactory = this.messageManagerItf
                .getChunkFactory(MessageExchangeChunk.class);
        this.protocolChunkFactory = this.messageManagerItf.getChunkFactory(ProtocolChunk.class);
        this.topologyService = topologyService;

        exportIdChunks = new HashMap<String, Map<String, ExportIdentifierChunk>>();

        this.log.end();
    }

    /**
     * Add a chunk corresponding to the specified protocol including the
     * destination of the message
     * 
     * @param message
     * @param host
     * @param port
     * @param protocol
     * @return
     * @throws TransportException
     */
    public Message addChunkDestinationAndProtocol(Message message, String containerName,
            String transportQOS) throws TransportException {
        this.log.start();

        if (!exportIdChunks.containsKey(containerName)) {
            ContainerConfiguration containerConfiguration = null;
            try {
                containerConfiguration = this.topologyService
                        .getContainerConfiguration(containerName);
            } catch (final PetalsException e) {
                throw new TransportException("Cannot get the information for the container "
                        + containerName);
            }
            this.createChunkDestination(containerConfiguration);
        }
        final Map<String, ExportIdentifierChunk> containerExportIdChunks = exportIdChunks
                .get(containerName);

        if (transportQOS == null) {
            transportQOS = defaultQOS;
        }

        String protocol;
        if (Transporter.FAST_POLICY.equals(transportQOS)) {
            protocol = "tcp";
        } else {
            protocol = "ssl";
        }

        this.log.debug("Use Protocol Chunk : " + protocol);

        // add the destination
        ExportIdentifierChunk eic = containerExportIdChunks.get(protocol);
        eic = this.messageManagerItf.cloneChunk(eic);
        this.messageManagerItf.addOrReplaceChunk(message, DESTINATION_CHUNK_NAME, eic);

        // add the protocol
        final ProtocolChunk protocolChunk = this.messageManagerItf
                .createChunk(this.protocolChunkFactory);
        protocolChunk.setProtocol(protocol);
        this.messageManagerItf
                .addOrReplaceChunk(message, ProtocolChunk.DEFAULT_NAME, protocolChunk);

        this.log.end();
        return message;
    }

    /**
     * 
     * 
     * @param containerName
     * @return
     * @throws TransportException
     */
    protected void createChunkDestination(ContainerConfiguration containerConfiguration)
            throws TransportException {
        this.log.start();

        final Map<String, ExportIdentifierChunk> containerExportIdChunks = new HashMap<String, ExportIdentifierChunk>();
        ExportIdentifierChunk eic = null;

        final String host = containerConfiguration.getHost();
        int port = containerConfiguration.getDreamSSLPort();
        if (port != 0) {
            final IPExportIdentifier exportIdentifier = new IPExportIdentifier(host, port);
            eic = this.messageManagerItf.createChunk(this.exportIdChunkFactory);
            eic.setExportIdentifier(exportIdentifier);
            containerExportIdChunks.put("ssl", eic);
            this.log.debug("SSL Export identifier " + host + ":" + port);
        }

        port = containerConfiguration.getDreamTCPPort();
        if (port != 0) {
            final IPExportIdentifier exportIdentifier = new IPExportIdentifier(host, port);
            eic = this.messageManagerItf.createChunk(this.exportIdChunkFactory);
            eic.setExportIdentifier(exportIdentifier);
            containerExportIdChunks.put("tcp", eic);
            this.log.end("TCP/IP Export identifier " + host + ":" + port);
        }

        exportIdChunks.put(containerConfiguration.getName(), containerExportIdChunks);
    }

    /**
     * Transform a JBI message into a message formatted for a specific router.
     * 
     * @param jbi
     *            the message to transform
     * @param parameter
     *            map of specific needed objects required by the transporter.
     * @return the message formatted for the transporter
     * @throws JBIException
     *             thrown if an error occurs during serialization
     */
    public MessageExchangeImpl dream2jbi(Message dreamMessage) throws JBIException {
        this.log.start();

        try {
            final MessageExchangeChunk mec = this.messageManagerItf.getChunk(dreamMessage,
                    MessageExchangeChunk.DEFAULT_NAME);
            if (mec == null) {
                throw new JBIException("MessageExchangeChunk not found !");
            } else {
                this.log.end();
                return mec.getMessageExchange();
            }
        } catch (final InvalidReference ir) {
            throw new JBIException(ir);
        }
    }

    /**
     * Transform a transporter format message into a JBI message.
     * 
     * @param messageFromTransporter
     *            message with a specific format related to the transporter
     * @param stuffs
     *            map of specific needed.
     * @return a jbi {@link MessageExchangeImpl message}
     * @throws TransportException
     *             thrown if an error occurs during serialization
     */
    public Message jbi2dream(MessageExchangeImpl jbi) throws TransportException {
        this.log.start();

        final Message dreamMessage = this.messageManagerItf.createMessage();
        try {
            final MessageExchangeChunk mec = this.messageManagerItf
                    .createChunk(this.messageExchangeChunkFactory);
            mec.setMessageExchange(jbi);
            this.messageManagerItf.addChunk(dreamMessage, MessageExchangeChunk.DEFAULT_NAME, mec);
            this.log.end();
            return dreamMessage;
        } catch (final InvalidReference ir) {
            throw new TransportException(ir);
        } catch (final ChunkAlreadyExistException caee) {
            throw new TransportException(caee);
        }
    }

    /**
     * Setup the Dream Transporter. Some configuration of DREAM is dynamic and
     * must be done before starting the DREAM transporter composite.
     * 
     * @throws PetalsException
     */
    public static final void setupDreamTransporter(ContentController petalsContentController,
            ContainerConfiguration containerConfiguration) throws PetalsException {
        final Map<Object, Object> parameters = new HashMap<Object, Object>();

        parameters.put("nbMaxSession", Integer.toString(containerConfiguration
                .getDreamSessionCacheSize()));

        // Create the DREAM TCP protocol if necessary
        if (containerConfiguration.getDreamTCPPort() != 0) {
            parameters.put("tcpPort", Integer.toString(containerConfiguration.getDreamTCPPort()));
            parameters.put("tcpHostName", containerConfiguration.getHost());
            parameters.put("connectTimeout", Integer.toString(containerConfiguration
                    .getDreamConnectionTimeout()));
            parameters.put("connectRetry", Integer.toString(containerConfiguration
                    .getDreamConnectionRetry()));

        }
        if (containerConfiguration.getDreamSSLPort() != 0) {
            parameters.put("sslPort", Integer.toString(containerConfiguration.getDreamSSLPort()));
            parameters.put("sslHostName", containerConfiguration.getHost());
            parameters.put("keyStoreFile", containerConfiguration.getSSLKeystore());
            parameters.put("keyStorePassword", containerConfiguration.getSSLKeystorePassword());
            parameters.put("keyPassword", containerConfiguration.getSSLKeyPassword());
            parameters.put("trustStoreFile", containerConfiguration.getSSLTruststore());
            parameters.put("trustStorePassword", containerConfiguration.getSSLTruststorePassword());
        }

        String dreamCompositeADLName;
        if (containerConfiguration.getDreamTCPPort() != 0) {
            if (containerConfiguration.getDreamSSLPort() != 0) {
                dreamCompositeADLName = "DreamTransportProtocolTCPandSSL";
                defaultQOS = Transporter.FAST_POLICY;
            } else {
                dreamCompositeADLName = "DreamTransportProtocolTCP";
                defaultQOS = Transporter.FAST_POLICY;
            }
        } else {
            dreamCompositeADLName = "DreamTransportProtocolSSL";
            defaultQOS = Transporter.SECURE_POLICY;
        }

        try {
            final Component dreamComponent = FractalHelper.createNewComponent(
                    dreamCompositeADLName, parameters);

            final NameController dreamNC = Fractal.getNameController(dreamComponent);
            dreamNC.setFcName(FractalHelper.DREAMTRANSPORTER_COMPONENT);

            final Component transporterComposite = FractalHelper.getRecursiveComponentByName(
                    petalsContentController, FractalHelper.TRANSPORTER_COMPOSITE);
            final Component dispatcherComponent = FractalHelper.getRecursiveComponentByName(
                    petalsContentController, FractalHelper.DISPATCHER_COMPONENT);

            final ContentController transporterCC = Fractal
                    .getContentController(transporterComposite);
            final BindingController dreamBC = Fractal.getBindingController(dreamComponent);
            transporterCC.addFcSubComponent(dreamComponent);
            dreamBC.bindFc("topology", transporterCC.getFcInternalInterface("topology"));
            dreamBC.bindFc("dispatcher", dispatcherComponent.getFcInterface("dispatcher"));

            final BindingController dispatcherBC = Fractal
                    .getBindingController(dispatcherComponent);
            dispatcherBC
                    .bindFc("delegated-service-dream", dreamComponent.getFcInterface("service"));

        } catch (final NoSuchInterfaceException e) {
            throw new PetalsException("Error while configuring Dream transporter", e);
        } catch (final SecurityException e) {
            throw new PetalsException("Error while configuring Dream transporter", e);
        } catch (final IllegalArgumentException e) {
            throw new PetalsException("Error while configuring Dream transporter", e);
        } catch (final IllegalContentException e) {
            throw new PetalsException("Error while configuring Dream transporter", e);
        } catch (final IllegalLifeCycleException e) {
            throw new PetalsException("Error while configuring Dream transporter", e);
        } catch (final IllegalBindingException e) {
            throw new PetalsException("Error while configuring Dream transporter", e);
        } catch (final ADLException e) {
            throw new PetalsException("Error while configuring Dream transporter", e);
        }
    }
}
