/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 Fossil E-Commerce, http://www.fossilec.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: NioUtil.java 10233 2009-02-23 15:30:55Z rnaudin $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.util;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.channels.Channels;
import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;
import java.nio.channels.OverlappingFileLockException;
import java.nio.channels.ReadableByteChannel;

/**
 * @version $Rev: 10233 $ $Date: 2009-02-23 16:30:55 +0100 (Mon, 23 Feb 2009) $
 * @since Petals 1.0
 * @author <a href="mailto:rmarins@fossilec.com">Rafael Marins</a>
 */
public final class NioUtil {

    /**
     * Size of buffer in bytes used during some operations.
     */
    private static final int BUFFER_SIZE = 4096;

    /**
     * Do not construct.
     */
    private NioUtil() {
        super();
    }

    /**
     * Transfer bytes blocks using direct buffer memory allocation. Relies in
     * use the faster Java New I/O instead of traditional byte array move.
     * 
     * @param is
     *            the {@link InputStream} to copy
     * @param aFile
     *            the output {@link File}
     * @throws IOException
     *             if errors occured
     */
    public static final void copyStreamToFile(InputStream is, File aFile)
            throws IOException {
        ReadableByteChannel inputChannel = Channels.newChannel(is);
        FileOutputStream fos = new FileOutputStream(aFile);
        FileChannel outputChannel = fos.getChannel();
        try {
            ByteBuffer buffer = ByteBuffer.allocate(BUFFER_SIZE);
            int readBytes;
            while ((readBytes = inputChannel.read(buffer)) != -1) {
                if (readBytes > 0) {
                    buffer.flip();
                    outputChannel.write(buffer);
                    buffer.clear();
                }
            }
        } finally {
            inputChannel.close();
            outputChannel.close();
            fos.close();
        }
    }
    
    /**
     * check if the file can be read
     * 
     * @param file
     * @return
     */
    public static final boolean checkAccess(File file) {
        boolean result = false;
        FileChannel channel = null;
        FileLock lock = null;
        try {
            // try to access the file to test if it can be read
            channel = new RandomAccessFile(file, "rw").getChannel();
            // Use the file channel to create a lock on the file.
            // This method blocks until it can retrieve the lock.
            lock = channel.lock();
            lock.release();
            result = true;
        } catch (OverlappingFileLockException e) {
            // do nothing
        } catch (IOException e) {
            // do nothing
        } finally {
            if (channel != null) {
                try {
                    channel.close();
                } catch (IOException e1) {
                    // do nothing
                }

            }
        }
        return result;
    }

}
